# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn

from mmrazor.registry import MODELS


@MODELS.register_module()
class OFDLoss(nn.Module):
    """A Comprehensive Overhaul of Feature Distillation
    https://sites.google.com/view/byeongho-heo/overhaul.

    The partial L2loss, only calculating loss when
    `out_s > out_t` or `out_t > 0`.

    Args:
        loss_weight (float, optional): loss weight. Defaults to 1.0.
        mul_factor (float, optional): multiply factor. Defaults to 1000.
    """

    def __init__(self,
                 loss_weight: float = 1.0,
                 mul_factor: float = 1000.) -> None:
        super(OFDLoss, self).__init__()
        self.loss_weight = loss_weight
        self.mul_factor = mul_factor

    def forward_train(self, s_feature: torch.Tensor,
                      t_feature: torch.Tensor) -> torch.Tensor:
        """forward func for training.

        Args:
            s_feature (torch.Tensor): student's feature
            t_feature (torch.Tensor): teacher's feature

        Returns:
            torch.Tensor: loss
        """
        bsz = s_feature.shape[0]
        loss = torch.nn.functional.mse_loss(
            s_feature, t_feature, reduction='none')
        loss = loss * ((s_feature > t_feature) | (t_feature > 0)).float()
        return loss.sum() / bsz / self.mul_factor

    def forward(self, s_feature: torch.Tensor,
                t_feature: torch.Tensor) -> torch.Tensor:
        """forward func.

        Args:
            s_feature (torch.Tensor): student's feature
            t_feature (torch.Tensor): teacher's feature

        Returns:
            torch.Tensor: loss
        """
        return self.loss_weight * self.forward_train(s_feature, t_feature)



# Copyright (c) OpenMMLab. All rights reserved.
from .dynamic_conv import BigNasConv2d, DynamicConv2d, OFAConv2d
from .dynamic_linear import DynamicLinear
from .dynamic_mixins import (DynamicBatchNormMixin, DynamicChannelMixin,
                             DynamicLinearMixin, DynamicMixin)
from .dynamic_norm import (DynamicBatchNorm1d, DynamicBatchNorm2d,
                           DynamicBatchNorm3d)

__all__ = [
    'BigNasConv2d', 'DynamicConv2d', 'OFAConv2d', 'DynamicLinear',
    'DynamicBatchNorm1d', 'DynamicBatchNorm2d', 'DynamicBatchNorm3d',
    'DynamicMixin', 'DynamicChannelMixin', 'DynamicBatchNormMixin',
    'DynamicLinearMixin'
]



# Copyright (c) OpenMMLab. All rights reserved.
import logging

from mmengine import fileio
from mmengine.logging import print_log
from torch import nn

from mmrazor.utils import FixMutable, ValidFixMutable


def _dynamic_to_static(model: nn.Module) -> None:
    # Avoid circular import
    from mmrazor.models.architectures.dynamic_ops.bricks import DynamicMixin

    def traverse_children(module: nn.Module) -> None:
        # TODO
        # dynamicop must have no dynamic child
        for name, child in module.named_children():
            if isinstance(child, DynamicMixin):
                setattr(module, name, child.to_static_op())
            else:
                traverse_children(child)

    if isinstance(model, DynamicMixin):
        raise RuntimeError('Root model can not be dynamic op.')

    traverse_children(model)


def load_fix_subnet(model: nn.Module,
                    fix_mutable: ValidFixMutable,
                    prefix: str = '') -> None:
    """Load fix subnet."""
    if isinstance(fix_mutable, str):
        fix_mutable = fileio.load(fix_mutable)
    if not isinstance(fix_mutable, dict):
        raise TypeError('fix_mutable should be a `str` or `dict`'
                        f'but got {type(fix_mutable)}')

    from mmrazor.models.architectures.dynamic_ops.bricks import DynamicMixin
    if isinstance(model, DynamicMixin):
        raise RuntimeError('Root model can not be dynamic op.')

    # Avoid circular import
    from mmrazor.models.mutables import DerivedMutable
    from mmrazor.models.mutables.base_mutable import BaseMutable

    for name, module in model.named_modules():
        # The format of `chosen`` is different for each type of mutable.
        # In the corresponding mutable, it will check whether the `chosen`
        # format is correct.
        if isinstance(module, BaseMutable):
            if getattr(module, 'alias', None):
                alias = module.alias
                assert alias in fix_mutable, \
                    f'The alias {alias} is not in fix_modules, ' \
                    'please check your `fix_mutable`.'
                chosen = fix_mutable.get(alias, None)
            else:
                mutable_name = name.lstrip(prefix)
                if mutable_name not in fix_mutable and \
                        not isinstance(module, DerivedMutable):
                    raise RuntimeError(
                        f'The module name {mutable_name} is not in '
                        'fix_mutable, please check your `fix_mutable`.')
                chosen = fix_mutable.get(mutable_name, None)
            module.fix_chosen(chosen)

    # convert dynamic op to static op
    _dynamic_to_static(model)


def export_fix_subnet(model: nn.Module,
                      dump_derived_mutable: bool = False) -> FixMutable:
    """Export subnet that can be loaded by :func:`load_fix_subnet`."""
    if dump_derived_mutable:
        print_log(
            'Trying to dump information of all derived mutables, '
            'this might harm readability of the exported configurations.',
            level=logging.WARNING)

    # Avoid circular import
    from mmrazor.models.mutables import DerivedMutable
    from mmrazor.models.mutables.base_mutable import BaseMutable

    fix_subnet = dict()
    for name, module in model.named_modules():
        if isinstance(module, BaseMutable):
            if isinstance(module, DerivedMutable) and not dump_derived_mutable:
                continue

            assert not module.is_fixed
            if module.alias:
                fix_subnet[module.alias] = module.dump_chosen()
            else:
                fix_subnet[name] = module.dump_chosen()

    return fix_subnet



# Copyright (c) OpenMMLab. All rights reserved.
from .derived_mutable import DerivedMutable
from .mutable_channel import (MutableChannel, OneShotMutableChannel,
                              SlimmableMutableChannel)
from .mutable_module import (DiffChoiceRoute, DiffMutableModule, DiffMutableOP,
                             OneShotMutableModule, OneShotMutableOP)
from .mutable_value import MutableValue, OneShotMutableValue

__all__ = [
    'OneShotMutableOP', 'OneShotMutableModule', 'DiffMutableOP',
    'DiffChoiceRoute', 'DiffMutableModule', 'OneShotMutableChannel',
    'SlimmableMutableChannel', 'MutableChannel', 'DerivedMutable',
    'MutableValue', 'OneShotMutableValue'
]



# Copyright (c) OpenMMLab. All rights reserved.
import random
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union

import numpy as np
import torch.nn as nn
from torch import Tensor

from mmrazor.registry import MODELS
from ..base_mutable import CHOICE_TYPE, CHOSEN_TYPE
from .mutable_module import MutableModule


class OneShotMutableModule(MutableModule[CHOICE_TYPE, CHOSEN_TYPE]):
    """Base class for one shot mutable module. A base type of ``MUTABLES`` for
    single path supernet such as Single Path One Shot.

    All subclass should implement the following APIs:

    - ``sample_choice()``
    - ``forward_fixed()``
    - ``forward_all()``
    - ``forward_choice()``

    Args:
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.

    Note:
        :meth:`forward_all` is called when calculating FLOPs.
    """

    def forward(self, x: Any) -> Any:
        """Calls either :func:`forward_fixed` or :func:`forward_choice`
        depending on whether :func:`is_fixed` is ``True`` and whether
        :func:`current_choice` is None.

        Note:
            :meth:`forward_fixed` is called in `fixed` mode.
            :meth:`forward_all` is called in `unfixed` mode with
                :func:`current_choice` is None.
            :meth:`forward_choice` is called in `unfixed` mode with
                :func:`current_choice` is not None.

        Args:
            x (Any): input data for forward computation.
            choice (CHOICE_TYPE, optional): the chosen key in ``MUTABLE``.

        Returns:
            Any: the result of forward
        """
        if self.is_fixed:
            return self.forward_fixed(x)
        if self.current_choice is None:
            return self.forward_all(x)
        else:
            return self.forward_choice(x, choice=self.current_choice)

    @abstractmethod
    def sample_choice(self) -> CHOICE_TYPE:
        """Sample random choice.

        Returns:
            CHOICE_TYPE: the chosen key in ``MUTABLE``.
        """

    @abstractmethod
    def forward_fixed(self, x: Any) -> Any:
        """Forward with the fixed mutable.

        All subclasses must implement this method.
        """

    @abstractmethod
    def forward_all(self, x: Any) -> Any:
        """Forward all choices.

        All subclasses must implement this method.
        """

    @abstractmethod
    def forward_choice(self, x: Any, choice: CHOICE_TYPE) -> Any:
        """Forward with the unfixed mutable and current_choice is not None.

        All subclasses must implement this method.
        """


@MODELS.register_module()
class OneShotMutableOP(OneShotMutableModule[str, str]):
    """A type of ``MUTABLES`` for single path supernet, such as Single Path One
    Shot. In single path supernet, each choice block only has one choice
    invoked at the same time. A path is obtained by sampling all the choice
    blocks.

    Args:
        candidates (dict[str, dict]): the configs for the candidate
            operations.
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.

    Examples:
        >>> import torch
        >>> from mmrazor.models.mutables import OneShotMutableOP

        >>> candidates = nn.ModuleDict({
        ...     'conv3x3': nn.Conv2d(32, 32, 3, 1, 1),
        ...     'conv5x5': nn.Conv2d(32, 32, 5, 1, 2),
        ...     'conv7x7': nn.Conv2d(32, 32, 7, 1, 3)})

        >>> input = torch.randn(1, 32, 64, 64)
        >>> op = OneShotMutableOP(candidates)

        >>> op.choices
        ['conv3x3', 'conv5x5', 'conv7x7']
        >>> op.num_choices
        3
        >>> op.is_fixed
        False

        >>> op.current_choice = 'conv3x3'
        >>> unfix_output = op.forward(input)
        >>> torch.all(unfixed_output == candidates['conv3x3'](input))
        True

        >>> op.fix_chosen('conv3x3')
        >>> fix_output = op.forward(input)
        >>> torch.all(fix_output == unfix_output)
        True

        >>> op.choices
        ['conv3x3']
        >>> op.num_choices
        1
        >>> op.is_fixed
        True
    """

    def __init__(
        self,
        candidates: Union[Dict[str, Dict], nn.ModuleDict],
        module_kwargs: Optional[Dict[str, Dict]] = None,
        alias: Optional[str] = None,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(
            module_kwargs=module_kwargs, alias=alias, init_cfg=init_cfg)
        assert len(candidates) >= 1, \
            f'Number of candidate op must greater than 1, ' \
            f'but got: {len(candidates)}'

        self._chosen: Optional[str] = None
        if isinstance(candidates, dict):
            self._candidates = self._build_ops(candidates, self.module_kwargs)
        elif isinstance(candidates, nn.ModuleDict):
            self._candidates = candidates
        else:
            raise TypeError('candidata_ops should be a `dict` or '
                            f'`nn.ModuleDict` instance, but got '
                            f'{type(candidates)}')

        assert len(self._candidates) >= 1, \
            f'Number of candidate op must greater than or equal to 1, ' \
            f'but got {len(self._candidates)}'

    @staticmethod
    def _build_ops(
            candidates: Union[Dict[str, Dict], nn.ModuleDict],
            module_kwargs: Optional[Dict[str, Dict]] = None) -> nn.ModuleDict:
        """Build candidate operations based on choice configures.

        Args:
            candidates (dict[str, dict] | :obj:`nn.ModuleDict`): the configs
                for the candidate operations or nn.ModuleDict.
            module_kwargs (dict[str, dict], optional): Module initialization
                named arguments.

        Returns:
            ModuleDict (dict[str, Any], optional):  the key of ``ops`` is
                the name of each choice in configs and the value of ``ops``
                is the corresponding candidate operation.
        """
        if isinstance(candidates, nn.ModuleDict):
            return candidates

        ops = nn.ModuleDict()
        for name, op_cfg in candidates.items():
            assert name not in ops
            if module_kwargs is not None:
                op_cfg.update(module_kwargs)
            ops[name] = MODELS.build(op_cfg)
        return ops

    def forward_fixed(self, x: Any) -> Tensor:
        """Forward with the `fixed` mutable.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.

        Returns:
            Tensor: the result of forward the fixed operation.
        """
        return self._candidates[self._chosen](x)

    def forward_choice(self, x: Any, choice: str) -> Tensor:
        """Forward with the `unfixed` mutable and current choice is not None.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.
            choice (str): the chosen key in ``OneShotMutableOP``.

        Returns:
            Tensor: the result of forward the ``choice`` operation.
        """
        assert isinstance(choice, str) and choice in self.choices
        return self._candidates[choice](x)

    def forward_all(self, x: Any) -> Tensor:
        """Forward all choices. Used to calculate FLOPs.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.

        Returns:
            Tensor: the result of forward all of the ``choice`` operation.
        """
        outputs = list()
        for op in self._candidates.values():
            outputs.append(op(x))
        return sum(outputs)

    def fix_chosen(self, chosen: str) -> None:
        """Fix mutable with subnet config. This operation would convert
        `unfixed` mode to `fixed` mode. The :attr:`is_fixed` will be set to
        True and only the selected operations can be retained.

        Args:
            chosen (str): the chosen key in ``MUTABLE``. Defaults to None.
        """
        if self.is_fixed:
            raise AttributeError(
                'The mode of current MUTABLE is `fixed`. '
                'Please do not call `fix_chosen` function again.')

        for c in self.choices:
            if c != chosen:
                self._candidates.pop(c)

        self._chosen = chosen
        self.is_fixed = True

    def dump_chosen(self) -> str:
        assert self.current_choice is not None

        return self.current_choice

    def sample_choice(self) -> str:
        """uniform sampling."""
        return np.random.choice(self.choices, 1)[0]

    @property
    def choices(self) -> List[str]:
        """list: all choices. """
        return list(self._candidates.keys())

    @property
    def num_choices(self):
        return len(self.choices)


@MODELS.register_module()
class OneShotProbMutableOP(OneShotMutableOP):
    """Sampling candidate operation according to probability.

    Args:
        candidates (dict[str, dict]): the configs for the candidate
            operations.
        choice_probs (list): the probability of sampling each
            candidate operation.
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self,
                 candidates: Dict[str, Dict],
                 choice_probs: list = None,
                 module_kwargs: Optional[Dict[str, Dict]] = None,
                 alias: Optional[str] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(
            candidates=candidates,
            module_kwargs=module_kwargs,
            alias=alias,
            init_cfg=init_cfg)
        assert choice_probs is not None
        assert sum(choice_probs) - 1 < np.finfo(np.float64).eps, \
            f'Please make sure the sum of the {choice_probs} is 1.'
        self.choice_probs = choice_probs

    def sample_choice(self) -> str:
        """Sampling with probabilities."""
        assert len(self.choice_probs) == len(self._candidates.keys())
        choice = random.choices(
            self.choices, weights=self.choice_probs, k=1)[0]
        return choice



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Tuple

import torch
import torch.nn as nn
from mmengine.optim import OPTIMIZERS, OptimWrapper
from mmengine.runner import load_checkpoint

from mmrazor.models.utils import add_prefix, set_requires_grad
from mmrazor.registry import MODELS
from ...base import BaseAlgorithm


@MODELS.register_module()
class DataFreeDistillation(BaseAlgorithm):
    """Algorithm for data-free teacher-student distillation Typically, the
    teacher is a pretrained model and the student is a small model trained on
    the generator's output. The student is trained to mimic the behavior of the
    teacher. The generator is trained to generate images that are similar to
    the real images.

    Args:
        distiller (dict): The config dict for built distiller.
        generator_distiller (dict): The distiller collecting outputs & losses
            to update the generator.
        teachers (dict[str, dict]): The dict of config dict for teacher models
            and their ckpt_path (optional).
        generator (dictl): The config dict for built distiller generator.
        student_iter (int): The number of student steps in train_step().
            Defaults to 1.
        student_train_first (bool): Whether to train student in first place.
            Defaults to False.
    """

    def __init__(self,
                 distiller: dict,
                 generator_distiller: dict,
                 teachers: Dict[str, Dict[str, dict]],
                 generator: dict,
                 student_iter: int = 1,
                 student_train_first: bool = False,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.student_iter = student_iter
        self.student_train_first = student_train_first
        self.distiller = MODELS.build(distiller)
        self.generator_distiller = MODELS.build(generator_distiller)

        if not isinstance(teachers, Dict):
            raise TypeError('teacher should be a `dict` but got '
                            f'{type(teachers)}')

        self.teachers = nn.ModuleDict()
        for teacher_name, cfg in teachers.items():
            self.teachers[teacher_name] = MODELS.build(cfg['build_cfg'])
            if 'ckpt_path' in cfg:
                # avoid loaded parameters be overwritten
                self.teachers[teacher_name].init_weights()
                _ = load_checkpoint(self.teachers[teacher_name],
                                    cfg['ckpt_path'])
            self.teachers[teacher_name].eval()
            set_requires_grad(self.teachers[teacher_name], False)

        if not isinstance(generator, Dict):
            raise TypeError('generator should be a `dict` instance, but got '
                            f'{type(generator)}')
        self.generator = MODELS.build(generator)

        # In ``DataFreeDistiller``, the recorder manager is just
        # constructed, but not really initialized yet.
        self.distiller.prepare_from_student(self.student)
        self.distiller.prepare_from_teacher(self.teachers)
        self.generator_distiller.prepare_from_student(self.student)
        self.generator_distiller.prepare_from_teacher(self.teachers)

    @property
    def student(self) -> nn.Module:
        """Alias for ``architecture``."""
        return self.architecture

    def train_step(self, data: Dict[str, List[dict]],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:
        """Train step for DataFreeDistillation.

        Args:
            data (Dict[str, List[dict]]): Data sampled by dataloader.
            optim_wrapper (OptimWrapper): A wrapper of optimizer to
                update parameters.
        """
        log_vars = dict()
        for _, teacher in self.teachers.items():
            teacher.eval()

        if self.student_train_first:
            _, dis_log_vars = self.train_student(data,
                                                 optim_wrapper['architecture'])
            _, generator_loss_vars = self.train_generator(
                data, optim_wrapper['generator'])
        else:
            _, generator_loss_vars = self.train_generator(
                data, optim_wrapper['generator'])
            _, dis_log_vars = self.train_student(data,
                                                 optim_wrapper['architecture'])

        log_vars.update(dis_log_vars)
        log_vars.update(generator_loss_vars)
        return log_vars

    def train_student(
            self, data: Dict[str, List[dict]], optimizer: OPTIMIZERS
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Train step for the student model.

        Args:
            data (Dict[str, List[dict]]): Data sampled by dataloader.
            optimizer (OPTIMIZERS): The optimizer to update student.
        """
        log_vars = dict()
        batch_size = len(data['inputs'])

        for _ in range(self.student_iter):
            fakeimg_init = torch.randn(
                (batch_size, self.generator.module.latent_dim))
            fakeimg = self.generator(fakeimg_init, batch_size).detach()

            with optimizer.optim_context(self.student):
                pseudo_data = self.data_preprocessor(data, True)
                pseudo_data_samples = pseudo_data['data_samples']
                # recorde the needed information
                with self.distiller.student_recorders:
                    _ = self.student(fakeimg, pseudo_data_samples, mode='loss')
                with self.distiller.teacher_recorders, torch.no_grad():
                    for _, teacher in self.teachers.items():
                        _ = teacher(fakeimg, pseudo_data_samples, mode='loss')
                loss_distill = self.distiller.compute_distill_losses()

            distill_loss, distill_log_vars = self.parse_losses(loss_distill)
            optimizer.update_params(distill_loss)
        log_vars = dict(add_prefix(distill_log_vars, 'distill'))

        return distill_loss, log_vars

    def train_generator(
            self, data: Dict[str, List[dict]], optimizer: OPTIMIZERS
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Train step for the generator.

        Args:
            data (Dict[str, List[dict]]): Data sampled by dataloader.
            optimizer (OPTIMIZERS): The optimizer to update generator.
        """
        batch_size = len(data['inputs'])
        fakeimg_init = torch.randn(
            (batch_size, self.generator.module.latent_dim))
        fakeimg = self.generator(fakeimg_init, batch_size)

        with optimizer.optim_context(self.generator):
            pseudo_data = self.data_preprocessor(data, True)
            pseudo_data_samples = pseudo_data['data_samples']
            # recorde the needed information
            with self.generator_distiller.student_recorders:
                _ = self.student(fakeimg, pseudo_data_samples, mode='loss')
            with self.generator_distiller.teacher_recorders:
                for _, teacher in self.teachers.items():
                    _ = teacher(fakeimg, pseudo_data_samples, mode='loss')
            loss_generator = self.generator_distiller.compute_distill_losses()

        generator_loss, generator_loss_vars = self.parse_losses(loss_generator)
        optimizer.update_params(generator_loss)
        log_vars = dict(add_prefix(generator_loss_vars, 'generator'))

        return generator_loss, log_vars


@MODELS.register_module()
class DAFLDataFreeDistillation(DataFreeDistillation):

    def train_step(self, data: Dict[str, List[dict]],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:
        """DAFL train step.

        Args:
            data (Dict[str, List[dict]): Data sampled by dataloader.
            optim_wrapper (OptimWrapper): A wrapper of optimizer to
                update parameters.
        """
        log_vars = dict()
        batch_size = len(data['inputs'])

        for _, teacher in self.teachers.items():
            teacher.eval()

        # fakeimg initialization and revised by generator.
        fakeimg_init = torch.randn(
            (batch_size, self.generator.module.latent_dim))
        fakeimg = self.generator(fakeimg_init, batch_size)
        pseudo_data = self.data_preprocessor(data, True)
        pseudo_data_samples = pseudo_data['data_samples']

        with optim_wrapper['generator'].optim_context(self.generator):
            # recorde the needed information
            with self.generator_distiller.student_recorders:
                _ = self.student(fakeimg, pseudo_data_samples, mode='loss')
            with self.generator_distiller.teacher_recorders:
                for _, teacher in self.teachers.items():
                    _ = teacher(fakeimg, pseudo_data_samples, mode='loss')
            loss_generator = self.generator_distiller.compute_distill_losses()

        generator_loss, generator_loss_vars = self.parse_losses(loss_generator)
        log_vars.update(add_prefix(generator_loss_vars, 'generator'))

        with optim_wrapper['architecture'].optim_context(self.student):
            # recorde the needed information
            with self.distiller.student_recorders:
                _ = self.student(
                    fakeimg.detach(), pseudo_data_samples, mode='loss')
            with self.distiller.teacher_recorders, torch.no_grad():
                for _, teacher in self.teachers.items():
                    _ = teacher(
                        fakeimg.detach(), pseudo_data_samples, mode='loss')
            loss_distill = self.distiller.compute_distill_losses()

        distill_loss, distill_log_vars = self.parse_losses(loss_distill)
        log_vars.update(add_prefix(distill_log_vars, 'distill'))

        optim_wrapper['generator'].update_params(generator_loss)
        optim_wrapper['architecture'].update_params(distill_loss)

        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
import functools
from types import FunctionType
from typing import Callable

from mmengine.utils import import_modules_from_strings

from mmrazor.registry import TASK_UTILS
from .distill_delivery import DistillDelivery


@TASK_UTILS.register_module()
class FunctionOutputsDelivery(DistillDelivery):
    """Delivery for intermediate results which are ``FunctionType``'s outputs.

    Args:
        func_path (str): The name of the function whose output needs to be
            delivered.
        max_keep_data (int): The length limitation of the queue. Outputs from
            the source model are pushed in the queue in order.

    Notes:
        The form of `func_path` needs special attention. For example,
        `anchor_inside_flags` is a function in mmdetection to check whether the
        anchors are inside the border. This function is in
        `mmdet/core/anchor/utils.py` and used in
        `mmdet/models/dense_heads/anchor_head`. Then the `func_path` should be
        `mmdet.models.dense_heads.anchor_head.anchor_inside_flags` but not
        `mmdet.core.anchor.utils.anchor_inside_flags`.

    Examples:
        >>> # Below code in toy_module.py
        >>> import random
        >>> def toy_func():
        >>>     return random.randint(0, 1000)

        >>> # Below code in main.py
        >>> # Teacher and student both will execute toy_func.
        >>> # Now, we want to deliver outputs from the teacher to
        >>> # the student
        >>> import toy_module
        >>> delivery = FunctionOutputsDeliver(
        ...     max_keep_data=1, func_path='toy_module.toy_func')

        >>> delivery.override_data = False
        >>> with delivery:
        ...     output_teacher = toy_module.toy_func()

        >>> delivery.override_data = True
        >>> with delivery:
        ...     output_student = toy_module.toy_func()

        >>> output_teacher == output_student
        True

        >>> # If a function (method) is executed more than once during the
        >>> # forward of the source model, all the outputs of this function
        >>> # (method) will be used to override function (method) outputs from
        >>> # the target model.
        >>> delivery = FunctionOutputsDeliver(
        ...     max_keep_data=2, func_path='toy_module.toy_func')

        >>> delivery.override_data = False
        >>> with delivery:
        ...     output1_tea = toy_module.toy_func()
        ...     output2_tea = toy_module.toy_func()

        >>> delivery.override_data = True
        >>> with delivery:
        ...     output1_stu = toy_module.toy_func()
        ...     output2_stu = toy_module.toy_func()

        >>> output1_stu == output1_tea and output2_stu == output2_tea
        True
    """

    def __init__(self, func_path: str, max_keep_data: int):
        super().__init__(max_keep_data)

        self._check_valid_path(func_path)
        module_path = self._get_module_path(func_path)
        try:
            module = import_modules_from_strings(module_path)
        except ImportError:
            raise ImportError(f'{module_path} is not imported correctly.')
        self.module = module

        func_name = self._get_func_name(func_path)
        assert hasattr(module, func_name), \
            f'{func_name} is not in {module_path}.'
        self.func_name = func_name

        origin_func = getattr(module, func_name)
        if not isinstance(origin_func, FunctionType):
            raise TypeError(f'{func_name} should be a FunctionType '
                            f'instance, but got {type(origin_func)}')
        self.origin_func = origin_func

    @staticmethod
    def _check_valid_path(func_path: str) -> None:
        """Check if the `func_path` is valid."""
        if not isinstance(func_path, str):
            raise TypeError(f'func_path should be a FunctionType '
                            f'instance, but got {type(func_path)}')

        assert len(func_path.split('.')) > 1, \
            'func_path must have at least one `.`'

    @staticmethod
    def _get_func_name(func_path: str) -> str:
        """Get the function name according to `func_path`."""
        return func_path.split('.')[-1]

    @staticmethod
    def _get_module_path(func_path: str) -> str:
        """Get the module name according to `func_path`."""
        return '.'.join(func_path.split('.')[:-1])

    def __enter__(self) -> None:
        """Enter the context manager.

        Wrap the origin function.
        """
        wrapped_func = self.deliver_wrapper(self.origin_func)
        setattr(self.module, self.func_name, wrapped_func)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager.

        Reset the origin function.
        """
        setattr(self.module, self.func_name, self.origin_func)

    def deliver_wrapper(self, origin_func: Callable) -> Callable:
        """Wrap the specific function to make the intermediate results of the
        model can be delivered."""

        @functools.wraps(origin_func)
        def wrap_func(*args, **kwargs):

            if self.override_data:
                assert not self.data_queue.empty(), 'pop from an empty queue'
                outputs = self.data_queue.get()
            else:
                assert not self.data_queue.full(), 'push into an full queue'
                outputs = origin_func(*args, **kwargs)
                self.data_queue.put(outputs)
            return outputs

        return wrap_func



# Copyright (c) OpenMMLab. All rights reserved.
from .candidate import Candidates
from .fix_subnet import export_fix_subnet, load_fix_subnet

__all__ = ['load_fix_subnet', 'export_fix_subnet', 'Candidates']



_base_ = [
    'mmrazor::_base_/settings/imagenet_bs2048_autoslim_pil.py',
    'mmcls::_base_/models/mobilenet_v2_1x.py',
    'mmcls::_base_/default_runtime.py',
]

supernet = _base_.model
supernet.backbone.widen_factor = 1.5
supernet.head.in_channels = 1920

# !dataset config
# ==========================================================================
# data preprocessor
data_preprocessor = dict(
    type='ImgDataPreprocessor',
    # RGB format normalization parameters
    mean=[123.675, 116.28, 103.53],
    std=[58.395, 57.12, 57.375],
    # convert image from BGR to RGB
    bgr_to_rgb=True)

# !autoslim algorithm config
# ==========================================================================
channel_cfg_paths = [
    'https://download.openmmlab.com/mmrazor/v1/autoslim/autoslim_mbv2_subnet_8xb256_in1k_flops-220M_acc-71.4_20220715-9c288f3b_subnet_cfg.yaml',  # noqa: E501
    'https://download.openmmlab.com/mmrazor/v1/autoslim/autoslim_mbv2_subnet_8xb256_in1k_flops-320M_acc-72.73_20220715-9aa8f8ae_subnet_cfg.yaml',  # noqa: E501
    'https://download.openmmlab.com/mmrazor/v1/autoslim/autoslim_mbv2_subnet_8xb256_in1k_flops-530M_acc-74.23_20220715-aa8754fe_subnet_cfg.yaml'  # noqa: E501
]

model = dict(
    _delete_=True,
    _scope_='mmrazor',
    type='SlimmableNetwork',
    architecture=supernet,
    data_preprocessor=data_preprocessor,
    channel_cfg_paths=channel_cfg_paths,
    mutator=dict(
        type='SlimmableChannelMutator',
        mutable_cfg=dict(type='SlimmableMutableChannel'),
        tracer_cfg=dict(
            type='BackwardTracer',
            loss_calculator=dict(type='ImageClassifierPseudoLoss'))))

model_wrapper_cfg = dict(
    type='mmrazor.SlimmableNetworkDDP',
    broadcast_buffers=False,
    find_unused_parameters=True)

optim_wrapper = dict(accumulative_counts=len(channel_cfg_paths))

val_cfg = dict(type='mmrazor.SlimmableValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Tuple

import torch
from torch import nn

from mmrazor.models.utils import add_prefix
from mmrazor.registry import MODELS

try:
    from mmcls.evaluation import Accuracy
    from mmcls.models.heads import LinearClsHead
    from mmcls.structures import ClsDataSample
except ImportError:
    from mmrazor.utils import get_placeholder
    Accuracy = get_placeholder('mmcls')
    LinearClsHead = get_placeholder('mmcls')
    ClsDataSample = get_placeholder('mmcls')


@MODELS.register_module()
class DartsSubnetClsHead(LinearClsHead):

    def __init__(self, aux_in_channels, aux_loss, **kwargs):
        super(DartsSubnetClsHead, self).__init__(**kwargs)
        self.aux_linear = nn.Linear(aux_in_channels, self.num_classes)
        self.aux_loss_module = MODELS.build(aux_loss)

    def forward_aux(self, feats: Tuple[torch.Tensor]):

        aux_feat = feats[0]
        aux_cls_score = self.aux_linear(aux_feat)
        return aux_cls_score

    def _get_aux_loss(self, cls_score: torch.Tensor,
                      data_samples: List[ClsDataSample], **kwargs):
        """Unpack data samples and compute loss."""
        # Unpack data samples and pack targets
        if 'score' in data_samples[0].gt_label:
            # Batch augmentation may convert labels to one-hot format scores.
            target = torch.stack([i.gt_label.score for i in data_samples])
        else:
            target = torch.hstack([i.gt_label.label for i in data_samples])

        # compute loss
        losses = dict()
        loss = self.aux_loss_module(
            cls_score, target, avg_factor=cls_score.size(0), **kwargs)
        losses['loss'] = loss

        # compute accuracy
        if self.cal_acc:
            assert target.ndim == 1, 'If you enable batch augmentation ' \
                'like mixup during training, `cal_acc` is pointless.'
            acc = Accuracy.calculate(cls_score, target, topk=self.topk)
            losses.update(
                {f'accuracy_top-{k}': a
                 for k, a in zip(self.topk, acc)})

        return losses

    def loss(self, feats: Tuple[torch.Tensor],
             data_samples: List[ClsDataSample], **kwargs) -> dict:
        """Calculate losses from the classification score.

        Args:
            feats (tuple[Tensor]): The features extracted from the backbone.
                Multiple stage inputs are acceptable but only the last stage
                will be used to classify. The shape of every item should be
                ``(num_samples, num_classes)``.
            data_samples (List[ClsDataSample]): The annotation data of
                every samples.
            **kwargs: Other keyword arguments to forward the loss module.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        losses = super().loss(feats, data_samples, **kwargs)

        aux_cls_score = self.forward_aux(feats)
        aux_losses = self._get_aux_loss(aux_cls_score, data_samples, **kwargs)

        losses.update(add_prefix(aux_losses, 'aux_head.'))

        return losses



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Tuple

import torch
import torch.distributed as dist


class GatherTensors(torch.autograd.Function):
    """Gather tensors from all GPUS, supporting backward propagation.

    See more details in torch.distributed.all_gather and
    torch.distributed.all_reduce.
    """

    @staticmethod
    def forward(ctx: Any, input: torch.Tensor) -> Tuple[Any, ...]:
        """Forward function.

        It must accept a context ctx as the first argument.

        The context can be used to store tensors that can be then retrieved
        during the backward pass.

        Args:
            ctx (Any): Context to be used for forward propagation.
            input (torch.Tensor): Tensor to be broadcast from current process.
        """
        output = [
            torch.empty_like(input) for _ in range(dist.get_world_size())
        ]
        dist.all_gather(output, input)
        return tuple(output)

    @staticmethod
    def backward(ctx: Any, *grads: torch.Tensor) -> torch.Tensor:
        """Backward function.

        It must accept a context :attr:`ctx` as the first argument, followed by
        as many outputs did :func:`forward` return, and it should return as
        many tensors, as there were inputs to :func:`forward`. Each argument is
        the gradient w.r.t the given output, and each returned value should be
        the gradient w.r.t. the corresponding input.

        The context can be used to retrieve tensors saved during the forward
        pass. It also has an attribute :attr:`ctx.needs_input_grad` as a tuple
        of booleans representing whether each input needs gradient. E.g.,
        :func:`backward` will have ``ctx.needs_input_grad[0] = True`` if the
        first input to :func:`forward` needs gradient computated w.r.t. the
        output.

        Args:
            ctx (Any): Context to be used for forward propagation.
            grads (torch.Tensor): Grads to be merged from current process.
        """
        rank = dist.get_rank()
        merged = torch.stack(grads)
        dist.all_reduce(merged)
        return merged[rank]



# Copyright (c) OpenMMLab. All rights reserved.
from .image_classifier_loss_calculator import ImageClassifierPseudoLoss
from .single_stage_detector_loss_calculator import \
    SingleStageDetectorPseudoLoss

__all__ = ['ImageClassifierPseudoLoss', 'SingleStageDetectorPseudoLoss']



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Any, Dict, Optional, Sequence

import torch
from mmengine.hooks import Hook
from mmengine.registry import HOOKS
from mmengine.structures import BaseDataElement

from mmrazor.models.task_modules import ResourceEstimator

DATA_BATCH = Optional[Sequence[dict]]


@HOOKS.register_module()
class EstimateResourcesHook(Hook):
    """Estimate model resources periodically.

    Args:
        interval (int): The saving period. If ``by_epoch=True``, interval
            indicates epochs, otherwise it indicates iterations.
            Defaults to -1, which means "never".
        by_epoch (bool): Saving checkpoints by epoch or by iteration.
            Default to True.
        estimator_cfg (Dict[str, Any]): Used for building a resource estimator.
            Default to dict().

    Example:
    >>> add the `EstimatorResourcesHook` in custom_hooks as follows:
        custom_hooks = [
            dict(type='mmrazor.EstimateResourcesHook',
                 interval=1,
                 by_epoch=True,
                 estimator_cfg=dict(input_shape=(1, 3, 64, 64)))
        ]
    """
    out_dir: str

    priority = 'VERY_LOW'

    def __init__(self,
                 interval: int = -1,
                 by_epoch: bool = True,
                 estimator_cfg: Dict[str, Any] = dict(),
                 **kwargs) -> None:
        self.interval = interval
        self.by_epoch = by_epoch
        self.estimator = ResourceEstimator(**estimator_cfg)

    def after_val_epoch(self,
                        runner,
                        metrics: Optional[Dict[str, float]] = None) -> None:
        """Estimate model resources after every n val epochs.

        Args:
            runner (Runner): The runner of the training process.
        """
        if not self.by_epoch:
            return

        if self.every_n_epochs(runner, self.interval):
            self.estimate_resources(runner)

    def after_val_iter(self,
                       runner,
                       batch_idx: int,
                       data_batch: DATA_BATCH = None,
                       outputs: Optional[Sequence[BaseDataElement]] = None) \
            -> None:
        """Estimate model resources after every n val iters.

        Args:
            runner (Runner): The runner of the training process.
        """
        if self.by_epoch:
            return

        if self.every_n_train_iters(runner, self.interval):
            self.estimate_resources(runner)

    def estimate_resources(self, runner) -> None:
        """Estimate model resources: latency/flops/params."""
        model = runner.model.module if runner.distributed else runner.model

        # TODO confirm the state judgement.
        if hasattr(model, 'is_supernet') and model.is_supernet:
            model = self.export_subnet(model)

        resource_metrics = self.estimator.estimate(model)
        runner.logger.info(f'Estimate model resources: {resource_metrics}')

    def export_subnet(self, model) -> torch.nn.Module:
        """Export current best subnet.

        NOTE: This method is called when it comes to those NAS algorithms that
        require building a supernet for training.

        For those algorithms, measuring subnet resources is more meaningful
        than supernet during validation, therefore this method is required to
        get the current searched subnet from the supernet.
        """
        # Avoid circular import
        from mmrazor.models.mutables.base_mutable import BaseMutable
        from mmrazor.structures import load_fix_subnet

        # delete non-leaf tensor to get deepcopy(model).
        # TODO solve the hard case.
        for module in model.architecture.modules():
            if isinstance(module, BaseMutable):
                if hasattr(module, 'arch_weights'):
                    delattr(module, 'arch_weights')

        copied_model = copy.deepcopy(model)
        fix_mutable = copied_model.search_subnet()
        load_fix_subnet(copied_model, fix_mutable)

        return copied_model



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn

from mmrazor.registry import MODELS


@MODELS.register_module()
class L2Loss(nn.Module):
    """Calculate the two-norm loss between the two features.

    Args:
        loss_weight (float): Weight of loss. Defaults to 1.0.
        normalize (bool): Whether to normalize the feature. Defaults to True.
        mult (float): Multiplier for feature normalization. Defaults to 1.0.
        div_element (bool): Whether to divide the loss by element-wise.
            Defaults to False.
        dist (bool): Whether to conduct two-norm dist as torch.dist(p=2).
            Defaults to False.
    """

    def __init__(
        self,
        loss_weight: float = 1.0,
        normalize: bool = True,
        mult: float = 1.0,
        div_element: bool = False,
        dist: bool = False,
    ) -> None:
        super().__init__()
        self.loss_weight = loss_weight
        self.normalize = normalize
        self.mult = mult
        self.div_element = div_element
        self.dist = dist

    def forward(
        self,
        s_feature: torch.Tensor,
        t_feature: torch.Tensor,
    ) -> torch.Tensor:
        """Forward computation.

        Args:
            s_feature (torch.Tensor): The student model feature with
                shape (N, C, H, W) or shape (N, C).
            t_feature (torch.Tensor): The teacher model feature with
                shape (N, C, H, W) or shape (N, C).
        """
        if self.normalize:
            s_feature = self.normalize_feature(s_feature)
            t_feature = self.normalize_feature(t_feature)

        loss = torch.sum(torch.pow(torch.sub(s_feature, t_feature), 2))

        # Calculate l2_loss as dist.
        if self.dist:
            loss = torch.sqrt(loss)
        else:
            if self.div_element:
                loss = loss / s_feature.numel()
            else:
                loss = loss / s_feature.size(0)

        return self.loss_weight * loss

    def normalize_feature(self, feature: torch.Tensor) -> torch.Tensor:
        """Normalize the input feature.

        Args:
            feature (torch.Tensor): The student model feature with
                shape (N, C, H, W) or shape (N, C).
        """
        feature = feature.view(feature.size(0), -1)
        return feature / feature.norm(2, dim=1, keepdim=True) * self.mult



# Copyright (c) OpenMMLab. All rights reserved.
import functools
from types import FunctionType, ModuleType
from typing import Callable, List, Optional

from mmengine.utils import import_modules_from_strings
from torch import nn

from mmrazor.registry import TASK_UTILS
from .base_recorder import BaseRecorder


@TASK_UTILS.register_module()
class FunctionOutputsRecorder(BaseRecorder):
    """Recorder for intermediate results which are ``FunctionType``'s outputs.

     Notes:
        The form of `source` needs special attention. For example,
        `anchor_inside_flags` is a function in mmdetection to check whether the
        anchors are inside the border. This function is in
        `mmdet/core/anchor/utils.py` and used in
        `mmdet/models/dense_heads/anchor_head.py`. Then the source should be
        `mmdet.models.dense_heads.anchor_head.anchor_inside_flags` but not
        `mmdet.core.anchor.utils.anchor_inside_flags`.


    Examples:
        >>> # Below code in toy_module.py
        >>> import random
        >>> def toy_func():
        ...     return random.randint(0, 1000)
        >>> def toy_list_func():
        ...     return [random.randint(0,1000) for _ in range(3)]

        >>> # Below code in main.py
        >>> # Now, we want to get teacher's outputs by recorder.

        >>> import toy_module
        >>> r1 = FunctionOutputsRecorder('toy_module.toy_func')
        >>> r1.initialize()
        >>> with r1:
        ...     output_teacher1 = toy_module.toy_func()
        ...     output_teacher2 = toy_module.toy_func()
        ...     output_teacher3 = toy_module.toy_func()

        >>> r1.data_buffer
        [33, 41, 12]
        >>> recorder.get_record_data(record_idx=2)
        12
        >>> output_teacher1==33 and output_teacher2==41 and output_teacher3==41
        True

        >>> r2 = FunctionOutputsRecorder('toy_module.toy_list_func')
        >>> r2.initialize()
        >>> with r2:
        ...     output_teacher1 = toy_module.toy_list_func()
        ...     output_teacher2 = toy_module.toy_list_func()
        ...     output_teacher3 = toy_module.toy_list_func()

        >>> r2.data_buffer
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        >>> r2.get_record_data(record_idx=2, data_idx=2)
        9
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._check_valid_source(self.source)

        # import the function corrosponding module
        try:
            mod = import_modules_from_strings(self.module_string)
        except ImportError:
            raise ImportError(
                f'{self.module_string} is not imported correctly.')

        self.imported_module: ModuleType = mod

        assert hasattr(mod, self.func_name), \
            f'{self.func_name} is not in {self.module_string}.'

        origin_func = getattr(mod, self.func_name)
        if not isinstance(origin_func, FunctionType):
            raise TypeError(f'{self.func_name} should be a FunctionType '
                            f'instance, but got {type(origin_func)}')

        self.origin_func: Callable = origin_func

    @staticmethod
    def _check_valid_source(source):
        """Check if the source's format is valid."""
        if not isinstance(source, str):
            raise TypeError(f'source should be a str '
                            f'instance, but got {type(source)}')

        assert len(source.split('.')) > 1, \
            'source must have at least one `.`'

    @property
    def func_name(self):
        """Get the function name according to `func_path`."""
        return self.source.split('.')[-1]

    @property
    def module_string(self):
        """Get the module name according to `func_path`."""
        return '.'.join(self.source.split('.')[:-1])

    def prepare_from_model(self, model: Optional[nn.Module] = None) -> None:
        """The `model` is useless in `FunctionOutputsRecorder`."""
        pass

    def func_record_wrapper(self, origin_func: Callable,
                            data_buffer: List) -> Callable:
        """Save the function's outputs.

        Args:
            origin_func (FunctionType): The method whose outputs need to be
                recorded.
            buffer_key (str): The key of the function's outputs saved in
                ``data_buffer``.
        """

        @functools.wraps(origin_func)
        def wrap_func(*args, **kwargs):
            outputs = origin_func(*args, **kwargs)
            # assume a func execute N times, there will be N outputs need to
            # save.
            data_buffer.append(outputs)
            return outputs

        return wrap_func

    def __enter__(self):
        """Enter the context manager."""
        super().__enter__()

        mod = self.imported_module
        origin_func = self.origin_func
        # add record wrapper to origin function.
        record_func = self.func_record_wrapper(origin_func, self.data_buffer)

        assert hasattr(mod, self.func_name), \
            f'{self.func_name} is not in {self.module_string}.'

        # rewrite the origin function
        setattr(mod, self.func_name, record_func)

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        super().__exit__(exc_type, exc_value, traceback)

        mod = self.imported_module
        origin_func = self.origin_func

        assert hasattr(mod, self.func_name), \
            f'{self.func_name} is not in {self.module_string}.'

        # restore the origin function
        setattr(mod, self.func_name, origin_func)



# Copyright (c) OpenMMLab. All rights reserved.
from .configurable import (DAFLDataFreeDistillation, DataFreeDistillation,
                           FpnTeacherDistill, OverhaulFeatureDistillation,
                           SelfDistill, SingleTeacherDistill)

__all__ = [
    'SingleTeacherDistill', 'FpnTeacherDistill', 'SelfDistill',
    'DataFreeDistillation', 'DAFLDataFreeDistillation',
    'OverhaulFeatureDistillation'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Dict, List, Optional

import torch
import torch.nn as nn

from mmrazor.registry import MODELS
from ...mutables import DiffMutableModule
from .module_mutator import ModuleMutator


@MODELS.register_module()
class DiffModuleMutator(ModuleMutator):
    """Differentiable mutable based mutator.

    `DiffModuleMutator` is responsible for mutating `DiffMutableModule`,
    `DiffMutableOP`, `DiffChoiceRoute` and `GumbelChoiceRoute`.
    The architecture parameters of the mutables are retained
    in `DiffModuleMutator`.

    Args:
        custom_group (list[list[str]], optional): User-defined search groups.
            All searchable modules that are not in ``custom_group`` will be
            grouped separately.
    """

    def __init__(self,
                 custom_group: Optional[List[List[str]]] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(custom_group=custom_group, init_cfg=init_cfg)

    def build_arch_param(self, num_choices) -> nn.Parameter:
        """Build learnable architecture parameters."""
        return nn.Parameter(torch.randn(num_choices) * 1e-3)

    def prepare_from_supernet(self, supernet: nn.Module) -> None:
        """Inherit from ``BaseMutator``'s, generate `arch_params` in DARTS.

        Args:
            supernet (:obj:`torch.nn.Module`): The architecture to be used
                in your algorithm.
        """

        super().prepare_from_supernet(supernet)
        self.arch_params = self.build_arch_params()
        self.modify_supernet_forward(self.arch_params)

    def build_arch_params(self):
        """This function will build many arch params, which are generally used
        in differentiable search algorithms, such as Darts' series. Each
        group_id corresponds to an arch param, so the Mutables with the same
        group_id share the same arch param.

        Returns:
            torch.nn.ParameterDict: the arch params are got by `search_groups`.
        """

        arch_params = nn.ParameterDict()

        for group_id, modules in self.search_groups.items():
            group_arch_param = self.build_arch_param(modules[0].num_choices)
            arch_params[str(group_id)] = group_arch_param

        return arch_params

    def modify_supernet_forward(self, arch_params):
        """Modify the DiffMutableModule's default arch_param in forward.

        In MMRazor, the `arch_param` is along to `DiffModuleMutator`, while the
        `DiffMutableModule` needs `arch_param` in the forward. Here we use
        partial function to assign the corresponding `arch_param` to each
        `DiffMutableModule`.
        """

        for group_id, mutables in self.search_groups.items():
            for m in mutables:
                m.set_forward_args(arch_param=arch_params[str(group_id)])

    def sample_choices(self):
        """Sampling by search groups.

        The sampling result of the first mutable of each group is the sampling
        result of this group.

        Returns:
            Dict[int, Any]: Random choices dict.
        """

        choices = dict()
        for group_id, mutables in self.search_groups.items():
            arch_parm = self.arch_params[str(group_id)]
            choice = mutables[0].sample_choice(arch_parm)
            choices[group_id] = choice
        return choices

    def set_choices(self, choices: Dict[int, Any]) -> None:
        """Set mutables' current choice according to choices sample by
        :func:`sample_choices`.

        Args:
            choices (Dict[int, Any]): Choices dict. The key is group_id in
                search groups, and the value is the sampling results
                corresponding to this group.
        """
        for group_id, mutables in self.search_groups.items():
            choice = choices[group_id]
            for m in mutables:
                m.current_choice = choice

    @property
    def mutable_class_type(self):
        """Differentiable mutable class type.

        Returns:
            Type[DiffMutableModule]: Class type of differentiable mutable.
        """
        return DiffMutableModule



# Copyright (c) OpenMMLab. All rights reserved.
import sys
from functools import partial

import torch
import torch.nn as nn

from mmrazor.registry import TASK_UTILS


def get_model_complexity_info(model,
                              input_shape=(1, 3, 224, 224),
                              spec_modules=[],
                              disabled_counters=[],
                              print_per_layer_stat=False,
                              as_strings=False,
                              input_constructor=None,
                              flush=False,
                              ost=sys.stdout):
    """Get complexity information of a model. This method can calculate FLOPs
    and parameter counts of a model with corresponding input shape. It can also
    print complexity information for each layer in a model. Supported layers
    are listed as below:

        - Convolutions: ``nn.Conv1d``, ``nn.Conv2d``, ``nn.Conv3d``.
        - Activations: ``nn.ReLU``, ``nn.PReLU``, ``nn.ELU``, ``nn.LeakyReLU``,
            ``nn.ReLU6``.
        - Poolings: ``nn.MaxPool1d``, ``nn.MaxPool2d``, ``nn.MaxPool3d``,
            ``nn.AvgPool1d``, ``nn.AvgPool2d``, ``nn.AvgPool3d``,
            ``nn.AdaptiveMaxPool1d``, ``nn.AdaptiveMaxPool2d``,
            ``nn.AdaptiveMaxPool3d``, ``nn.AdaptiveAvgPool1d``,
            ``nn.AdaptiveAvgPool2d``, ``nn.AdaptiveAvgPool3d``.
        - BatchNorms: ``nn.BatchNorm1d``, ``nn.BatchNorm2d``,
            ``nn.BatchNorm3d``.
        - Linear: ``nn.Linear``.
        - Deconvolution: ``nn.ConvTranspose2d``.
        - Upsample: ``nn.Upsample``.

    Args:
        model (nn.Module): The model for complexity calculation.
        input_shape (tuple): Input shape (including batchsize) used for
            calculation. Default to (1, 3, 224, 224)
        spec_modules (list): A list that contains the names of several spec
            modules, which users want to get resources infos of them.
            e.g., ['backbone', 'head'], ['backbone.layer1']. Default to [].
        disabled_counters (list): One can limit which ops' spec would be
            calculated. Default to [].
        print_per_layer_stat (bool): Whether to print complexity information
            for each layer in a model. Default to True.
        as_strings (bool): Output FLOPs and params counts in a string form.
            Default to True.
        input_constructor (None | callable): If specified, it takes a callable
            method that generates input. otherwise, it will generate a random
            tensor with input shape to calculate FLOPs. Default to None.
        flush (bool): same as that in :func:`print`. Default to False.
        ost (stream): same as ``file`` param in :func:`print`.
            Default to sys.stdout.

    Returns:
        tuple[float | str] | dict[str, float]: If `as_strings` is set to True,
            it will return FLOPs and parameter counts in a string format.
            Otherwise, it will return those in a float number format.
            If len(spec_modules) > 0, it will return a resource info dict with
            FLOPs and parameter counts of each spec module in float format.
    """
    assert type(input_shape) is tuple
    assert len(input_shape) >= 1
    assert isinstance(model, nn.Module)
    flops_params_model = add_flops_params_counting_methods(model)
    flops_params_model.eval()
    flops_params_model.start_flops_params_count(disabled_counters)
    if input_constructor:
        input = input_constructor(input_shape)
        _ = flops_params_model(**input)
    else:
        try:
            batch = torch.ones(()).new_empty(
                tuple(input_shape),
                dtype=next(flops_params_model.parameters()).dtype,
                device=next(flops_params_model.parameters()).device)
        except StopIteration:
            # Avoid StopIteration for models which have no parameters,
            # like `nn.Relu()`, `nn.AvgPool2d`, etc.
            batch = torch.ones(()).new_empty(tuple(input_shape))

        _ = flops_params_model(batch)

    flops_count, params_count = \
        flops_params_model.compute_average_flops_params_cost()

    if print_per_layer_stat:
        print_model_with_flops_params(
            flops_params_model,
            flops_count,
            params_count,
            ost=ost,
            flush=flush)

    if len(spec_modules):
        module_names = [name for name, _ in flops_params_model.named_modules()]
        for module in spec_modules:
            assert module in module_names, \
                f'All modules in spec_modules should be in the measured ' \
                f'flops_params_model. Got module {module} in spec_modules.'
        spec_modules_resources = dict()
        accumulate_sub_module_flops_params(flops_params_model)
        for name, module in flops_params_model.named_modules():
            if name in spec_modules:
                spec_modules_resources[name] = dict()
                spec_modules_resources[name]['flops'] = module.__flops__
                spec_modules_resources[name]['params'] = module.__params__
                if as_strings:
                    spec_modules_resources[name]['flops'] = str(
                        params_units_convert(module.__flops__,
                                             'G')) + ' GFLOPs'
                    spec_modules_resources[name]['params'] = str(
                        params_units_convert(module.__params__, 'M')) + ' M'

    flops_params_model.stop_flops_params_count()

    if len(spec_modules):
        return spec_modules_resources

    if as_strings:
        flops_string = str(params_units_convert(flops_count, 'G')) + ' GFLOPs'
        params_string = str(params_units_convert(params_count, 'M')) + ' M'
        return flops_string, params_string

    return flops_count, params_count


def params_units_convert(num_params, units='M', precision=3):
    """Convert parameter number with units.

    Args:
        num_params (float): Parameter number to be converted.
        units (str | None): Converted FLOPs units. Options are None, 'M',
            'K' and ''. If set to None, it will automatically choose the most
            suitable unit for Parameter number. Default to None.
        precision (int): Digit number after the decimal point. Default to 2.

    Returns:
        str: The converted parameter number.

    Examples:
        >>> params_units_convert(1e9)
        '1000.0'
        >>> params_units_convert(2e5)
        '200.0'
        >>> params_units_convert(3e-9)
        '3e-09'
    """

    if units == 'G':
        return round(num_params / 10.**9, precision)
    elif units == 'M':
        return round(num_params / 10.**6, precision)
    elif units == 'K':
        return round(num_params / 10.**3, precision)
    else:
        raise ValueError(f'Unsupported units convert: {units}')


def print_model_with_flops_params(model,
                                  total_flops,
                                  total_params,
                                  units='G',
                                  precision=3,
                                  ost=sys.stdout,
                                  flush=False):
    """Print a model with FLOPs and Params for each layer.

    Args:
        model (nn.Module): The model to be printed.
        total_flops (float): Total FLOPs of the model.
        total_params (float): Total parameter counts of the model.
        units (str | None): Converted FLOPs units. Default to 'G'.
        precision (int): Digit number after the decimal point. Default to 3.
        ost (stream): same as `file` param in :func:`print`.
            Default to sys.stdout.
        flush (bool): same as that in :func:`print`. Default to False.

    Example:
        >>> class ExampleModel(nn.Module):
        >>> def __init__(self):
        >>>     super().__init__()
        >>>     self.conv1 = nn.Conv2d(3, 8, 3)
        >>>     self.conv2 = nn.Conv2d(8, 256, 3)
        >>>     self.conv3 = nn.Conv2d(256, 8, 3)
        >>>     self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        >>>     self.flatten = nn.Flatten()
        >>>     self.fc = nn.Linear(8, 1)
        >>> def forward(self, x):
        >>>     x = self.conv1(x)
        >>>     x = self.conv2(x)
        >>>     x = self.conv3(x)
        >>>     x = self.avg_pool(x)
        >>>     x = self.flatten(x)
        >>>     x = self.fc(x)
        >>>     return x
        >>> model = ExampleModel()
        >>> x = (3, 16, 16)
        to print the complexity information state for each layer, you can use
        >>> get_model_complexity_info(model, x)
        or directly use
        >>> print_model_with_flops_params(model, 4579784.0, 37361)
        ExampleModel(
          0.037 M, 100.000% Params, 0.005 GFLOPs, 100.000% FLOPs,
          (conv1): Conv2d(0.0 M, 0.600% Params, 0.0 GFLOPs, 0.959% FLOPs, 3, 8, kernel_size=(3, 3), stride=(1, 1))  # noqa: E501
          (conv2): Conv2d(0.019 M, 50.020% Params, 0.003 GFLOPs, 58.760% FLOPs, 8, 256, kernel_size=(3, 3), stride=(1, 1))
          (conv3): Conv2d(0.018 M, 49.356% Params, 0.002 GFLOPs, 40.264% FLOPs, 256, 8, kernel_size=(3, 3), stride=(1, 1))
          (avg_pool): AdaptiveAvgPool2d(0.0 M, 0.000% Params, 0.0 GFLOPs, 0.017% FLOPs, output_size=(1, 1))
          (flatten): Flatten(0.0 M, 0.000% Params, 0.0 GFLOPs, 0.000% FLOPs, )
          (fc): Linear(0.0 M, 0.024% Params, 0.0 GFLOPs, 0.000% FLOPs, in_features=8, out_features=1, bias=True)
        )
    """

    def accumulate_params(self):
        if is_supported_instance(self):
            return self.__params__
        else:
            sum = 0
            for m in self.children():
                sum += m.accumulate_params()
            return sum

    def accumulate_flops(self):
        if is_supported_instance(self):
            return self.__flops__ / model.__batch_counter__
        else:
            sum = 0
            for m in self.children():
                sum += m.accumulate_flops()
            return sum

    def flops_repr(self):
        accumulated_num_params = self.accumulate_params()
        accumulated_flops_cost = self.accumulate_flops()
        flops_string = str(
            params_units_convert(
                accumulated_flops_cost, units=units,
                precision=precision)) + ' ' + units + 'FLOPs'
        params_string = str(
            params_units_convert(
                accumulated_num_params, units='M', precision=precision)) + ' M'
        return ', '.join([
            params_string,
            '{:.3%} Params'.format(accumulated_num_params / total_params),
            flops_string,
            '{:.3%} FLOPs'.format(accumulated_flops_cost / total_flops),
            self.original_extra_repr()
        ])

    def add_extra_repr(m):
        m.accumulate_flops = accumulate_flops.__get__(m)
        m.accumulate_params = accumulate_params.__get__(m)
        flops_extra_repr = flops_repr.__get__(m)
        if m.extra_repr != flops_extra_repr:
            m.original_extra_repr = m.extra_repr
            m.extra_repr = flops_extra_repr
            assert m.extra_repr != m.original_extra_repr

    def del_extra_repr(m):
        if hasattr(m, 'original_extra_repr'):
            m.extra_repr = m.original_extra_repr
            del m.original_extra_repr
        if hasattr(m, 'accumulate_flops'):
            del m.accumulate_flops

    model.apply(add_extra_repr)
    print(model, file=ost, flush=flush)
    model.apply(del_extra_repr)


def accumulate_sub_module_flops_params(model):
    """Accumulate FLOPs and params for each module in the model. Each module in
    the model will have the `__flops__` and `__params__` parameters.

    Args:
        model (nn.Module): The model to be accumulated.
    """

    def accumulate_params(module):
        if is_supported_instance(module):
            return module.__params__
        else:
            sum = 0
            for m in module.children():
                sum += accumulate_params(m)
            return sum

    def accumulate_flops(module):
        if is_supported_instance(module):
            return module.__flops__ / model.__batch_counter__
        else:
            sum = 0
            for m in module.children():
                sum += accumulate_flops(m)
            return sum

    for module in model.modules():
        _flops = accumulate_flops(module)
        _params = accumulate_params(module)
        module.__flops__ = _flops
        module.__params__ = _params


def get_model_parameters_number(model):
    """Calculate parameter number of a model.

    Args:
        model (nn.module): The model for parameter number calculation.
    Returns:
        float: Parameter number of the model.
    """
    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    return num_params


def add_flops_params_counting_methods(net_main_module):
    # adding additional methods to the existing module object,
    # this is done this way so that each function has access to self object
    net_main_module.start_flops_params_count = start_flops_params_count.__get__(  # noqa: E501
        net_main_module)
    net_main_module.stop_flops_params_count = stop_flops_params_count.__get__(
        net_main_module)
    net_main_module.reset_flops_params_count = reset_flops_params_count.__get__(  # noqa: E501
        net_main_module)
    net_main_module.compute_average_flops_params_cost = compute_average_flops_params_cost.__get__(  # noqa: E501
        net_main_module)

    net_main_module.reset_flops_params_count()

    return net_main_module


def compute_average_flops_params_cost(self):
    """Compute average FLOPs and Params cost.

    A method to compute average FLOPs cost, which will be available after
    `add_flops_params_counting_methods()` is called on a desired net object.
    Returns:
        float: Current mean flops consumption per image.
    """
    batches_count = self.__batch_counter__
    flops_sum = 0
    params_sum = 0
    for module in self.modules():
        if is_supported_instance(module):
            flops_sum += module.__flops__
            params_sum += module.__params__
    return flops_sum / batches_count, params_sum


def start_flops_params_count(self, disabled_counters):
    """Activate the computation of mean flops and params consumption per image.

    A method to activate the computation of mean flops consumption per image.
    which will be available after ``add_flops_params_counting_methods()`` is
    called on a desired net object. It should be called before running the
    network.
    """
    add_batch_counter_hook_function(self)

    def add_flops_params_counter_hook_function(module):
        if is_supported_instance(module):
            if hasattr(module, '__flops_params_handle__'):
                return

            else:
                counter_type = get_counter_type(module)
                if (disabled_counters is None
                        or counter_type not in disabled_counters):
                    counter = TASK_UTILS.build(
                        dict(type=counter_type, _scope_='mmrazor'))
                    handle = module.register_forward_hook(
                        counter.add_count_hook)

                    module.__flops_params_handle__ = handle
                else:
                    return

    self.apply(partial(add_flops_params_counter_hook_function))


def stop_flops_params_count(self):
    """Stop computing the mean flops and params consumption per image.

    A method to stop computing the mean flops consumption per image, which will
    be available after ``add_flops_params_counting_methods()`` is called on a
    desired net object. It can be called to pause the computation whenever.
    """
    remove_batch_counter_hook_function(self)
    self.apply(remove_flops_params_counter_hook_function)


def reset_flops_params_count(self):
    """Reset statistics computed so far.

    A method to Reset computed statistics, which will be available after
    `add_flops_params_counting_methods()` is called on a desired net object.
    """
    add_batch_counter_variables_or_reset(self)
    self.apply(add_flops_params_counter_variable_or_reset)


# ---- Internal functions
def empty_flops_params_counter_hook(module, input, output):
    module.__flops__ += 0
    module.__params__ += 0


def add_batch_counter_variables_or_reset(module):

    module.__batch_counter__ = 0


def add_batch_counter_hook_function(module):
    if hasattr(module, '__batch_counter_handle__'):
        return

    handle = module.register_forward_hook(batch_counter_hook)
    module.__batch_counter_handle__ = handle


def batch_counter_hook(module, input, output):
    batch_size = 1
    if len(input) > 0:
        # Can have multiple inputs, getting the first one
        input = input[0]
        batch_size = len(input)
    else:
        pass
        print('Warning! No positional inputs found for a module, '
              'assuming batch size is 1.')
    module.__batch_counter__ += batch_size


def remove_batch_counter_hook_function(module):
    if hasattr(module, '__batch_counter_handle__'):
        module.__batch_counter_handle__.remove()
        del module.__batch_counter_handle__


def add_flops_params_counter_variable_or_reset(module):
    if is_supported_instance(module):
        if hasattr(module, '__flops__') or hasattr(module, '__params__'):
            print('Warning: variables __flops__ or __params__ are already '
                  'defined for the module' + type(module).__name__ +
                  ' ptflops can affect your code!')
        module.__flops__ = 0
        module.__params__ = 0


def get_counter_type(module):
    return module.__class__.__name__ + 'Counter'


def is_supported_instance(module):
    if get_counter_type(module) in TASK_UTILS._module_dict.keys():
        return True
    return False


def remove_flops_params_counter_hook_function(module):
    if hasattr(module, '__flops_params_handle__'):
        module.__flops_params_handle__.remove()
        del module.__flops_params_handle__
    if hasattr(module, '__flops__'):
        del module.__flops__
    if hasattr(module, '__params__'):
        del module.__params__



# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from collections import Counter
from typing import Dict, List, Optional, Type

from torch.nn import Module

from ..base_mutator import MUTABLE_TYPE, BaseMutator


class ModuleMutator(BaseMutator[MUTABLE_TYPE]):
    """The base class for mutable based mutator.

    All subclass should implement the following APIS:

    - ``mutable_class_type``

    Args:
        custom_group (list[list[str]], optional): User-defined search groups.
            All searchable modules that are not in ``custom_group`` will be
            grouped separately.
    """

    def __init__(self,
                 custom_group: Optional[List[List[str]]] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg)

        if custom_group is None:
            custom_group = []
        self._custom_group = custom_group
        self._search_groups: Optional[Dict[int, List[MUTABLE_TYPE]]] = None

    # TODO
    # should be a class property
    @property
    @abstractmethod
    def mutable_class_type(self) -> Type[MUTABLE_TYPE]:
        """Corresponding mutable class type.

        Returns:
            Type[MUTABLE_TYPE]: Mutable class type.
        """

    def prepare_from_supernet(self, supernet: Module) -> None:
        """Do some necessary preparations with supernet.

        Note:
            For mutable based mutator, we need to build search group first.

        Args:
            supernet (:obj:`torch.nn.Module`): The supernet to be searched
                in your algorithm.
        """
        self._build_search_groups(supernet)

    @property
    def search_groups(self) -> Dict[int, List[MUTABLE_TYPE]]:
        """Search group of supernet.

        Note:
            For mutable based mutator, the search group is composed of
            corresponding mutables.

        Raises:
            RuntimeError: Called before search group has been built.

        Returns:
            Dict[int, List[MUTABLE_TYPE]]: Search group.
        """
        if self._search_groups is None:
            raise RuntimeError(
                'Call `prepare_from_supernet` before access search group!')
        return self._search_groups

    def _build_name_mutable_mapping(
            self, supernet: Module) -> Dict[str, MUTABLE_TYPE]:
        """Mapping module name to mutable."""
        name2mutable: Dict[str, MUTABLE_TYPE] = dict()
        for name, module in supernet.named_modules():
            if isinstance(module, self.mutable_class_type):
                name2mutable[name] = module
        return name2mutable

    def _build_alias_names_mapping(self,
                                   supernet: Module) -> Dict[str, List[str]]:
        """Mapping alias to module names."""
        alias2mutable_names: Dict[str, List[str]] = dict()
        for name, module in supernet.named_modules():
            if isinstance(module, self.mutable_class_type):
                if module.alias is not None:
                    if module.alias not in alias2mutable_names:
                        alias2mutable_names[module.alias] = [name]
                    else:
                        alias2mutable_names[module.alias].append(name)

        return alias2mutable_names

    def _build_search_groups(self, supernet: Module) -> None:
        """Build search group with ``custom_group`` and ``alias``(see more
        information in :class:`BaseMutable`). Grouping by alias and module name
        are both supported.

        Note:
            Apart from user-defined search group, all other searchable
            modules(mutable) will be grouped separately.

            The main difference between using alias and module name for
            grouping is that the alias is One-to-Many while the module
            name is One-to-One.

            When using both alias and module name in `custom_group`, the
            priority of alias is higher than that of module name.

            If alias is set in `custom_group`, then its corresponding module
            name should not be in the `custom_group`.

            Moreover, there should be no duplicate keys in the `custom_group`.

        Example:
            >>> import torch
            >>> from mmrazor.models.mutables.diff_mutable import DiffMutableOP

            >>> # Assume that a toy model consists of three mutabels
            >>> # whose name are op1,op2,op3. The corresponding
            >>> # alias names of the three mutables are a1, a1, a2.
            >>> model = ToyModel()

            >>> # Using alias for grouping
            >>> mutator = DiffMutableOP(custom_group=[['a1'], ['a2']])
            >>> mutator.prepare_from_supernet(model)
            >>> mutator.search_groups
            {0: [op1, op2], 1: [op3]}

            >>> # Using module name for grouping
            >>> mutator = DiffMutableOP(custom_group=[['op1', 'op2'], ['op3']])
            >>> mutator.prepare_from_supernet(model)
            >>> mutator.search_groups
            {0: [op1, op2], 1: [op3]}

            >>> # Using both alias and module name for grouping
            >>> mutator = DiffMutableOP(custom_group=[['a2'], ['op2']])
            >>> mutator.prepare_from_supernet(model)
            >>> # The last operation would be grouped
            >>> mutator.search_groups
            {0: [op3], 1: [op2], 2: [op1]}


        Args:
            supernet (:obj:`torch.nn.Module`): The supernet to be searched
                in your algorithm.
        """
        name2mutable = self._build_name_mutable_mapping(supernet)
        alias2mutable_names = self._build_alias_names_mapping(supernet)

        # Check whether the custom group is valid
        if len(self._custom_group) > 0:
            self._check_valid_groups(alias2mutable_names, name2mutable,
                                     self._custom_group)

        # Construct search_groups based on user-defined group
        search_groups: Dict[int, List[MUTABLE_TYPE]] = dict()

        current_group_nums = 0
        grouped_mutable_names: List[str] = list()
        grouped_alias: List[str] = list()
        for group in self._custom_group:
            group_mutables = list()
            for item in group:
                if item in alias2mutable_names:
                    # if the item is from alias name
                    mutable_names: List[str] = alias2mutable_names[item]
                    grouped_alias.append(item)
                    group_mutables.extend(
                        [name2mutable[n] for n in mutable_names])
                    grouped_mutable_names.extend(mutable_names)
                else:
                    # if the item is in name2mutable
                    group_mutables.append(name2mutable[item])
                    grouped_mutable_names.append(item)

            search_groups[current_group_nums] = group_mutables
            current_group_nums += 1

        # Construct search_groups based on alias
        for alias, mutable_names in alias2mutable_names.items():
            if alias not in grouped_alias:
                # Check whether all current names are already grouped
                flag_all_grouped = True
                for mutable_name in mutable_names:
                    if mutable_name not in grouped_mutable_names:
                        flag_all_grouped = False

                # If not all mutables are already grouped
                if not flag_all_grouped:
                    search_groups[current_group_nums] = []
                    for mutable_name in mutable_names:
                        if mutable_name not in grouped_mutable_names:
                            search_groups[current_group_nums].append(
                                name2mutable[mutable_name])
                            grouped_mutable_names.append(mutable_name)
                    current_group_nums += 1

        # check whether all the mutable objects are in the search_groups
        for name, module in supernet.named_modules():
            if isinstance(module, self.mutable_class_type):
                if name in grouped_mutable_names:
                    continue
                else:
                    search_groups[current_group_nums] = [module]
                    current_group_nums += 1

        grouped_counter = Counter(grouped_mutable_names)

        # find duplicate keys
        duplicate_keys = list()
        for key, count in grouped_counter.items():
            if count > 1:
                duplicate_keys.append(key)

        assert len(grouped_mutable_names) == len(
            list(set(grouped_mutable_names))), \
            'There are duplicate keys in grouped mutable names. ' \
            f'The duplicate keys are {duplicate_keys}. ' \
            'Please check if there are duplicate keys in the `custom_group`.'

        self._search_groups = search_groups

    def _check_valid_groups(self, alias2mutable_names: Dict[str, List[str]],
                            name2mutable: Dict[str, MUTABLE_TYPE],
                            custom_group: List[List[str]]) -> None:

        aliases = [*alias2mutable_names.keys()]
        module_names = [*name2mutable.keys()]

        # check if all keys are legal
        expanded_custom_group: List[str] = [
            _ for group in custom_group for _ in group
        ]
        legal_keys: List[str] = [*aliases, *module_names]

        for key in expanded_custom_group:
            if key not in legal_keys:
                raise AssertionError(
                    f'The key: {key} in `custom_group` is not legal. '
                    f'Legal keys are: {legal_keys}. '
                    'Make sure that the keys are either alias or mutable name')

        # when the mutable has alias attribute, the corresponding module
        # name should not be used in `custom_group`.
        used_aliases = list()
        for group in custom_group:
            for key in group:
                if key in aliases:
                    used_aliases.append(key)

        for alias_key in used_aliases:
            mutable_names: List = alias2mutable_names[alias_key]
            # check whether module name is in custom group
            for mutable_name in mutable_names:
                if mutable_name in expanded_custom_group:
                    raise AssertionError(
                        f'When a mutable is set alias attribute :{alias_key},'
                        f'the corresponding module name {mutable_name} should '
                        f'not be used in `custom_group` {custom_group}.')



# Copyright (c) OpenMMLab. All rights reserved.
from .diff_mutable_module import (DiffChoiceRoute, DiffMutableModule,
                                  DiffMutableOP)
from .mutable_module import MutableModule
from .one_shot_mutable_module import OneShotMutableModule, OneShotMutableOP

__all__ = [
    'DiffMutableModule', 'DiffMutableOP', 'DiffChoiceRoute',
    'OneShotMutableOP', 'OneShotMutableModule', 'MutableModule'
]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import shutil
from pathlib import Path
from typing import Union

import torch
from mmengine import digit_version


def parse_args():
    parser = argparse.ArgumentParser(
        description='Process a checkpoint to be published')
    parser.add_argument('ckpt', help='input checkpoint filename', type=str)
    parser.add_argument('--model-name', help='model(config) name', type=str)
    parser.add_argument('--timestamp', help='training timestamp', type=str)
    parser.add_argument('--out-dir', help='output dir', type=str)
    args = parser.parse_args()
    return args


def cal_file_sha256(file_path: Union[str, Path]) -> str:
    import hashlib

    BLOCKSIZE = 65536
    sha256_hash = hashlib.sha256()

    with open(file_path, 'rb') as f:
        block = f.read(BLOCKSIZE)
        while block:
            sha256_hash.update(block)
            block = f.read(BLOCKSIZE)

    return sha256_hash.hexdigest()


def process_checkpoint(ckpt_path_str: str, model_name: str, timestamp: str,
                       out_dir_str: str) -> None:

    ckpt_path = Path(ckpt_path_str)
    work_dir = ckpt_path.parent

    out_dir: Path = Path(out_dir_str)
    out_dir.mkdir(parents=True, exist_ok=True)

    tmp_ckpt_path = out_dir / 'tmp.pth'

    checkpoint = torch.load(ckpt_path, map_location='cpu')
    # remove optimizer for smaller file size
    if 'optimizer' in checkpoint:
        del checkpoint['optimizer']
    # remove message_hub for smaller file size
    if 'message_hub' in checkpoint:
        del checkpoint['message_hub']
    # remove param_schedulers for smaller file size
    if 'param_schedulers' in checkpoint:
        del checkpoint['param_schedulers']

    # if it is necessary to remove some sensitive data in checkpoint['meta'],
    # add the code here.
    if digit_version(torch.__version__) >= digit_version('1.6'):
        torch.save(
            checkpoint, tmp_ckpt_path, _use_new_zipfile_serialization=False)
    else:
        torch.save(checkpoint, tmp_ckpt_path)

    sha = cal_file_sha256(tmp_ckpt_path)
    save_ckpt_path = f'{out_dir}/{model_name}_{timestamp}-{sha[:8]}.pth'
    tmp_ckpt_path.rename(save_ckpt_path)
    print(f'Successfully generated the publish-ckpt as {save_ckpt_path}.')

    log_path = work_dir / timestamp / f'{timestamp}.log'
    save_log_path = f'{out_dir}/{model_name}_{timestamp}-{sha[:8]}.log'
    shutil.copy(str(log_path), str(save_log_path))
    print(f'Successfully generated the publish-log as {save_log_path}.')

    log_path = work_dir / timestamp / f'{timestamp}.log'
    json_path = work_dir / timestamp / f'vis_data/{timestamp}.json'
    save_json_path = f'{out_dir}/{model_name}_{timestamp}-{sha[:8]}.json'
    shutil.copy(str(json_path), str(save_json_path))
    print(f'Successfully generated the publish-log as {save_json_path}.')


def main():
    args = parse_args()
    process_checkpoint(args.ckpt, args.model_name, args.timestamp,
                       args.out_dir)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, List, Optional, Sequence, Tuple, Union

from mmcv.cnn import ConvModule
from mmengine.model import Sequential
from torch import Tensor
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.registry import MODELS

try:
    from mmcls.models.backbones.base_backbone import BaseBackbone
    from mmcls.models.utils import make_divisible
except ImportError:
    from mmrazor.utils import get_placeholder
    BaseBackbone = get_placeholder('mmcls')
    make_divisible = get_placeholder('mmcls')


@MODELS.register_module()
class SearchableMobileNet(BaseBackbone):
    """Searchable MobileNet backbone.

    Args:
        arch_setting (list[list]): Architecture settings.
        first_channels (int): Channel width of first ConvModule. Default: 32.
        last_channels (int): Channel width of last ConvModule. Default: 1200.
        widen_factor (float): Width multiplier, multiply number of
            channels in each layer by this amount. Default: 1.0.
        out_indices (Sequence[int]): Output from which stages.
            Default: (7, ).
        frozen_stages (int): Stages to be frozen (all param fixed).
            Default: -1, which means not freezing any parameters.
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU6').
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        init_cfg (dict | list[dict], optional): initialization configuration
            dict to define initializer. OpenMMLab has implemented
            6 initializers, including ``Constant``, ``Xavier``, ``Normal``,
            ``Uniform``, ``Kaiming``, and ``Pretrained``.

    Excamples:
        >>> mutable_cfg = dict(
        ...     type='OneShotMutableOP',
        ...     candidates=dict(
        ...         mb_k3e1=dict(
        ...             type='MBBlock',
        ...             kernel_size=3,
        ...             expand_ratio=1,
        ...             norm_cfg=dict(type='BN'),
        ...             act_cfg=dict(type='ReLU6'))))
        >>> arch_setting = [
        ...     # Parameters to build layers. 4 parameters are needed to
        ...     # construct a layer, from left to right:
        ...     # channel, num_blocks, stride, mutable cfg.
        ...     [16, 1, 1, mutable_cfg],
        ...     [24, 2, 2, mutable_cfg],
        ...     [32, 3, 2, mutable_cfg],
        ...     [64, 4, 2, mutable_cfg],
        ...     [96, 3, 1, mutable_cfg],
        ...     [160, 3, 2, mutable_cfg],
        ...     [320, 1, 1, mutable_cfg]
        ... ]
        >>> model = SearchableMobileNet(arch_setting=arch_setting)
    """

    def __init__(
        self,
        arch_setting: List[List],
        first_channels: int = 32,
        last_channels: int = 1280,
        widen_factor: float = 1.,
        out_indices: Sequence[int] = (7, ),
        frozen_stages: int = -1,
        conv_cfg: Optional[Dict] = None,
        norm_cfg: Dict = dict(type='BN'),
        act_cfg: Dict = dict(type='ReLU6'),
        norm_eval: bool = False,
        with_cp: bool = False,
        init_cfg: Optional[Union[Dict, List[Dict]]] = [
            dict(type='Kaiming', layer=['Conv2d']),
            dict(type='Constant', val=1, layer=['_BatchNorm', 'GroupNorm'])
        ]
    ) -> None:
        for index in out_indices:
            if index not in range(8):
                raise ValueError('the item in out_indices must in '
                                 f'range(0, 8). But received {index}')

        if frozen_stages not in range(-1, 8):
            raise ValueError('frozen_stages must be in range(-1, 8). '
                             f'But received {frozen_stages}')

        super().__init__(init_cfg)

        self.arch_setting = arch_setting
        self.widen_factor = widen_factor
        self.out_indices = out_indices
        self.frozen_stages = frozen_stages
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.norm_eval = norm_eval
        self.with_cp = with_cp

        self.in_channels = make_divisible(first_channels * widen_factor, 8)

        self.conv1 = ConvModule(
            in_channels=3,
            out_channels=self.in_channels,
            kernel_size=3,
            stride=2,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

        self.layers = []

        for i, layer_cfg in enumerate(arch_setting):
            channel, num_blocks, stride, mutable_cfg = layer_cfg
            out_channels = make_divisible(channel * widen_factor, 8)
            inverted_res_layer = self._make_layer(
                out_channels=out_channels,
                num_blocks=num_blocks,
                stride=stride,
                mutable_cfg=copy.deepcopy(mutable_cfg))
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, inverted_res_layer)
            self.layers.append(layer_name)

        if widen_factor > 1.0:
            self.out_channel = int(last_channels * widen_factor)
        else:
            self.out_channel = last_channels

        layer = ConvModule(
            in_channels=self.in_channels,
            out_channels=self.out_channel,
            kernel_size=1,
            stride=1,
            padding=0,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

        self.add_module('conv2', layer)
        self.layers.append('conv2')

    def _make_layer(self, out_channels: int, num_blocks: int, stride: int,
                    mutable_cfg: Dict) -> Sequential:
        """Stack mutable blocks to build a layer for SearchableMobileNet.

        Note:
            Here we use ``module_kwargs`` to pass dynamic parameters such as
            ``in_channels``, ``out_channels`` and ``stride``
            to build the mutable.

        Args:
            out_channels (int): out_channels of block.
            num_blocks (int): number of blocks.
            stride (int): stride of the first block.
            mutable_cfg (dict): Config of mutable.

        Returns:
            mmengine.model.Sequential: The layer made.
        """
        layers = []
        for i in range(num_blocks):
            if i >= 1:
                stride = 1

            mutable_cfg.update(
                module_kwargs=dict(
                    in_channels=self.in_channels,
                    out_channels=out_channels,
                    stride=stride))
            layers.append(MODELS.build(mutable_cfg))

            self.in_channels = out_channels

        return Sequential(*layers)

    def forward(self, x: Tensor) -> Tuple[Tensor, ...]:
        """Forward computation.

        Args:
            x (tensor): x contains input data for forward computation.
        """
        x = self.conv1(x)

        outs = []
        for i, layer_name in enumerate(self.layers):
            layer = getattr(self, layer_name)
            x = layer(x)
            if i in self.out_indices:
                outs.append(x)

        return tuple(outs)

    def _freeze_stages(self) -> None:
        """Freeze params not to update in the specified stages."""
        if self.frozen_stages >= 0:
            for param in self.conv1.parameters():
                param.requires_grad = False
        for i in range(1, self.frozen_stages + 1):
            layer = getattr(self, f'layer{i}')
            layer.eval()
            for param in layer.parameters():
                param.requires_grad = False

    def train(self, mode: bool = True) -> None:
        """Set module status before forward computation."""
        super().train(mode)

        self._freeze_stages()
        if mode and self.norm_eval:
            for m in self.modules():
                if isinstance(m, _BatchNorm):
                    m.eval()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Tuple

from torch import nn

from mmrazor.registry import TASK_UTILS
from .module_outputs_recorder import ModuleOutputsRecorder


@TASK_UTILS.register_module()
class ModuleInputsRecorder(ModuleOutputsRecorder):
    """Recorder for intermediate results which are Pytorch moudle's inputs."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward_hook(self, module: nn.Module, inputs: Tuple,
                     outputs: Any) -> None:
        """Save the module's forward input.

        Args:
            module (:obj:`torch.nn.Module`): The module to register hook.
            inputs (tuple): The input of the module.
            outputs : The output of the module.
        """
        if self.recording:
            self.data_buffer.append(inputs)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch.nn as nn
from mmcv.cnn import ConvModule

from mmrazor.registry import MODELS
from .base import BaseOP

try:
    from mmcls.models.utils import SELayer
except ImportError:
    from mmrazor.utils import get_placeholder
    SELayer = get_placeholder('mmcls')


@MODELS.register_module()
class ConvBnAct(BaseOP):
    """ConvBnAct block from timm.

    Args:
        in_channels (int): number of in channels.
        out_channels (int): number of out channels.
        kernel_size (int): kernel size of convolution.
        stride (int, optional): stride of convolution. Defaults to 1.
        dilation (int, optional): dilation rate of convolution. Defaults to 1.
        padding (int, optional): padding size of convolution. Defaults to 0.
        skip (bool, optional): whether using skip connect. Defaults to False.
        conv_cfg (Optional[dict], optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (Dict, optional): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (Dict, optional):Config dict for activation layer.
            Default: dict(type='ReLU').
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 stride: int = 1,
                 dilation: int = 1,
                 padding: int = 0,
                 skip: bool = False,
                 conv_cfg: Optional[dict] = None,
                 se_cfg: Dict = None,
                 norm_cfg: Dict = dict(type='BN'),
                 act_cfg: Dict = dict(type='ReLU')):
        super().__init__(
            in_channels=in_channels, out_channels=out_channels, stride=stride)
        self.has_residual = skip and stride == 1 \
            and in_channels == out_channels
        self.with_se = se_cfg is not None

        if self.with_se:
            assert isinstance(se_cfg, dict)
            self.se = SELayer(self.out_channels, **se_cfg)

        self.convModule = ConvModule(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            padding=padding,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

    def forward(self, x):
        """Forward function."""
        shortcut = x
        x = self.convModule(x)
        if self.has_residual:
            x += shortcut
        return x


@MODELS.register_module()
class DepthwiseSeparableConv(BaseOP):
    """DepthwiseSeparable block Used for DS convs in MobileNet-V1 and in the
    place of IR blocks that have no expansion (factor of 1.0). This is an
    alternative to having a IR with an optional first pw conv.

    Args:
        in_channels (int): number of in channels.
        out_channels (int): number of out channels.
        dw_kernel_size (int, optional): the kernel size of depth-wise
            convolution. Defaults to 3.
        stride (int, optional): stride of convolution.
            Defaults to 1.
        dilation (int, optional): dilation rate of convolution.
            Defaults to 1.
        noskip (bool, optional): whether use skip connection.
            Defaults to False.
        pw_kernel_size (int, optional): kernel size of point wise convolution.
            Defaults to 1.
        pw_act (bool, optional): whether using activation in point-wise
            convolution. Defaults to False.
        se_cfg (Dict, optional): _description_. Defaults to None.
        conv_cfg (Optional[dict], optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (Dict, optional): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (Dict, optional):Config dict for activation layer.
            Default: dict(type='ReLU').
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 dw_kernel_size: int = 3,
                 stride: int = 1,
                 dilation: int = 1,
                 noskip: bool = False,
                 pw_kernel_size: int = 1,
                 pw_act: bool = False,
                 conv_cfg: Optional[dict] = None,
                 se_cfg: Dict = None,
                 norm_cfg: Dict = dict(type='BN'),
                 act_cfg: Dict = dict(type='ReLU')):

        super().__init__(
            in_channels=in_channels, out_channels=out_channels, stride=stride)
        self.has_residual = (stride == 1
                             and in_channels == out_channels) and not noskip
        self.has_pw_act = pw_act  # activation after point-wise conv

        self.se_cfg = se_cfg

        self.conv_dw = ConvModule(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=dw_kernel_size,
            stride=stride,
            dilation=dilation,
            padding=dw_kernel_size // 2,
            groups=in_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )

        # Squeeze-and-excitation
        self.se = SELayer(out_channels, **
                          se_cfg) if self.se_cfg else nn.Identity()

        self.conv_pw = ConvModule(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=pw_kernel_size,
            padding=pw_kernel_size // 2,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg if self.has_pw_act else None,
        )

    def forward(self, x):
        shortcut = x
        x = self.conv_dw(x)
        x = self.se(x)
        x = self.conv_pw(x)
        if self.has_residual:
            x += shortcut
        return x



# Copyright (c) OpenMMLab. All rights reserved.
from .function_outputs_recorder import FunctionOutputsRecorder
from .method_outputs_recorder import MethodOutputsRecorder
from .module_inputs_recorder import ModuleInputsRecorder
from .module_outputs_recorder import ModuleOutputsRecorder
from .param_recorder import ParameterRecorder
from .recorder_manager import RecorderManager

__all__ = [
    'FunctionOutputsRecorder', 'MethodOutputsRecorder',
    'ModuleOutputsRecorder', 'ParameterRecorder', 'RecorderManager',
    'ModuleInputsRecorder'
]



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, Optional

from mmrazor.registry import TASK_UTILS
from .distill_delivery import DistillDelivery

SUPPORT_DELIVERIES = ['FunctionOutputs', 'MethodOutputs']


class DistillDeliveryManager:
    """Various types deliveries' manager. The ``DistillDeliveryManager`` is
    also a context manager, managing various types of deliveries.

    When entering the ``DistillDeliveryManager``, all deliveries managed by it
    will be started.

    Notes:
        DistillDelivery is a context manager used to override function(method)
        outputs during teacher(student) forward.

    Args:
        deliveries (dict): Configs of all deliveries.

    Examples:
        >>> from mmcls.models.utils import Augments

        >>> augments_cfg = dict(
        ...     type='BatchMixup', alpha=1., num_classes=10, prob=1.0)
        >>> augments = Augments(augments_cfg)
        >>> imgs = torch.randn(2, 3, 32, 32)
        >>> label = torch.randint(0, 10, (2,))

        >>> # Without ``MethodOutputsDelivery``, outputs of the teacher and
        >>> # the student are different.
        >>> imgs_tea, label_tea = augments(imgs, label)
        >>> imgs_stu, label_stu = augments(imgs, label)
        >>> torch.equal(label_tea, label_stu)
        False
        >>> torch.equal(imgs_tea, imgs_stu)
        False

        >>> distill_deliveries = ConfigDict(
        ...     aug=dict(type='MethodOutputs', max_keep_data=1,
        ...             method_path='mmcls.models.utils.Augments.__call__'))
        >>> manager = DistillDeliveryManager(distill_deliveries)

        >>> manager.override_data = False
        >>> with manager:
        ...     imgs_tea, label_tea = augments(imgs, label)

        >>> manager.override_data = True
        >>> with manager:
        ...     imgs_stu, label_stu = augments(imgs, label)

        >>> torch.equal(label_tea, label_stu)
        True
        >>> torch.equal(imgs_tea, imgs_stu)
        True
    """

    def __init__(self, deliveries: Optional[Dict[str, Dict]] = None) -> None:

        self._deliveries: Dict[str, DistillDelivery] = dict()
        if deliveries:
            for delivery_name, delivery_cfg in deliveries.items():
                delivery_cfg_ = copy.deepcopy(delivery_cfg)
                delivery_type_ = delivery_cfg_.get('type', '')
                assert isinstance(delivery_type_, str)
                assert delivery_type_ in SUPPORT_DELIVERIES

                delivery_type_ = delivery_type_ + 'Delivery'
                delivery_cfg_.update(dict(type=delivery_type_))

                delivery = TASK_UTILS.build(delivery_cfg_)
                self.deliveries[delivery_name] = delivery

        self._override_data = False

    @property
    def deliveries(self) -> Dict[str, DistillDelivery]:
        """dict: all deliveries."""
        return self._deliveries

    @property
    def override_data(self) -> bool:
        """bool: indicate whether to override the data with the recorded data.
        """
        return self._override_data

    @override_data.setter
    def override_data(self, override: bool) -> None:
        """Set the override_data property to all the deliveries.

        If the `override_data` of a delivery is False, the delivery will
        record the origin data.

        If the `override_data` of a delivery is True, the delivery will
        override the origin data with the recorded data.
        """
        self._override_data = override
        for delivery in self.deliveries.values():
            delivery.override_data = override

    def __enter__(self) -> None:
        """Enter the context manager."""
        for delivery in self.deliveries.values():
            delivery.__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager."""
        for delivery in self.deliveries.values():
            delivery.__exit__(exc_type, exc_value, traceback)



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np

from mmrazor.registry import TASK_UTILS
from ..flops_params_counter import get_model_parameters_number
from .base_counter import BaseCounter


class BNCounter(BaseCounter):
    """FLOPs/params counter for BatchNormalization series."""

    @staticmethod
    def add_count_hook(module, input, output):
        input = input[0]
        batch_flops = np.prod(input.shape)
        if getattr(module, 'affine', False):
            batch_flops *= 2
        module.__flops__ += int(batch_flops)
        module.__params__ += get_model_parameters_number(module)


@TASK_UTILS.register_module()
class BatchNorm1dCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class BatchNorm2dCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class BatchNorm3dCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class InstanceNorm1dCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class InstanceNorm2dCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class InstanceNorm3dCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class LayerNormCounter(BNCounter):
    pass


@TASK_UTILS.register_module()
class GroupNormCounter(BNCounter):
    pass



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from typing import Any, List, Optional

from torch import nn


class BaseRecorder(metaclass=ABCMeta):
    """Base class for recorders.

    Recorder is a context manager used to record various intermediate results
    during the model forward. It can be used in distillation algorithm and
    can also be used to obtain some specific data for visual analysis.

    In MMRazor, there will be different types of recorders to obtain different
    types of intermediate results. They can be used in combination with the
    ``RecorderManager``.

    Note:
        The recorder will be lazily initialized in the ``RecorderManager`` by
        default. If you want to use the recorder without the
        ``RecorderManager``, you need to initialize it first.
    """

    def __init__(self, source: str) -> None:

        self._source = source
        # Intermediate results are recorded in dictionary format according
        # to the data source.
        # One data source may generate multiple records, which need to be
        # recorded through list.
        self._data_buffer: List = list()
        # Before using the recorder for the first time, it needs to be
        # initialized.
        self._initialized = False

    @property
    def source(self) -> str:
        """str: source of recorded data."""
        return self._source

    @property
    def data_buffer(self) -> List:
        """list: data buffer."""
        return self._data_buffer

    @abstractmethod
    def prepare_from_model(self, model: Optional[nn.Module] = None) -> None:
        """Make the intermediate results of the model can be record."""

    def initialize(self, model: Optional[nn.Module] = None) -> None:
        """Init the recorder.

        Args:
            model (nn.Module): The model which need to record intermediate
                results.
        """
        self.prepare_from_model(model)
        self._initialized = True

    def get_record_data(self,
                        record_idx: int = 0,
                        data_idx: Optional[int] = None) -> Any:
        """Get data from ``data_buffer``.

        Args:
            record_idx (int): The index of the record saved in
                ``data_buffer``. If a source is executed N times during
                forward, there will be N records in ``data_buffer``.
            data_index (int, optional):  The index of target data in
                a record. A record may be a tuple or a list, if data_idx is
                None, the whole list or tuple is returned. Defaults to None.

        Returns:
            Any: The type of the return value is undefined, and different
                source data may have different types.
        """
        assert record_idx < len(self._data_buffer), \
            'record_idx is illegal. The length of data_buffer is ' \
            f'{len(self._data_buffer)}, but record_idx is ' \
            f'{record_idx}.'

        record = self._data_buffer[record_idx]

        if data_idx is None:
            target_data = record
        else:
            if isinstance(record, (list, tuple)):
                assert data_idx < len(record), \
                    'data_idx is illegal. The length of record is ' \
                    f'{len(record)}, but data_idx is {data_idx}.'
                target_data = record[data_idx]
            else:
                raise TypeError('When data_idx is not None, record should be '
                                'a list or tuple instance, but got '
                                f'{type(record)}.')

        return target_data

    def reset_data_buffer(self) -> None:
        """Clear data in data_buffer."""

        self._data_buffer = list()

    def __enter__(self):
        """Enter the context manager."""

        assert self._initialized, \
            'The recorder will be initialized in the RecorderManager by '\
            'default. If you want to use the recorder without the '\
            'RecorderManager, you need to initialize it first.'

        self.reset_data_buffer()

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, List, Optional, Sequence, Tuple, Union

import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import ModuleList, Sequential
from mmengine.model.weight_init import constant_init, normal_init
from torch import Tensor
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.registry import MODELS

try:
    from mmcls.models.backbones.base_backbone import BaseBackbone
except ImportError:
    from mmrazor.utils import get_placeholder
    BaseBackbone = get_placeholder('mmcls')


@MODELS.register_module()
class SearchableShuffleNetV2(BaseBackbone):
    """Based on ShuffleNetV2 backbone.

    Args:
        arch_setting (list[list]): Architecture settings.
        stem_multiplier (int): Stem multiplier - adjusts the number of
            channels in the first layer. Default: 1.
        widen_factor (float): Width multiplier - adjusts the number of
            channels in each layer by this amount. Default: 1.0.
        out_indices (Sequence[int]): Output from which stages.
            Default: (4, ).
        frozen_stages (int): Stages to be frozen (all param fixed).
            Default: -1, which means not freezing any parameters.
        with_last_layer (bool): Whether is last layer.
            Default: True, which means not need to add `Placeholder``.
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU').
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        init_cfg (dict | list[dict], optional): initialization configuration
            dict to define initializer. OpenMMLab has implemented
            6 initializers, including ``Constant``, ``Xavier``, ``Normal``,
            ``Uniform``, ``Kaiming``, and ``Pretrained``.

    Excamples:
        >>> mutable_cfg = dict(
        ...     type='OneShotMutableOP',
        ...     candidates=dict(
        ...         shuffle_3x3=dict(
        ...             type='ShuffleBlock',
        ...             kernel_size=3,
        ...             norm_cfg=dict(type='BN'))))
        >>> arch_setting = [
        ...     # Parameters to build layers. 3 parameters are needed to
        ...     # construct a layer, from left to right:
        ...     # channel, num_blocks, mutable cfg.
        ...     [64, 4, mutable_cfg],
        ...     [160, 4, mutable_cfg],
        ...     [320, 8, mutable_cfg],
        ...     [640, 4, mutable_cfg]
        ... ]
        >>> model = SearchableShuffleNetV2(arch_setting=arch_setting)
    """

    def __init__(self,
                 arch_setting: List[List],
                 stem_multiplier: int = 1,
                 widen_factor: float = 1.0,
                 out_indices: Sequence[int] = (4, ),
                 frozen_stages: int = -1,
                 with_last_layer: bool = True,
                 conv_cfg: Optional[Dict] = None,
                 norm_cfg: Dict = dict(type='BN'),
                 act_cfg: Dict = dict(type='ReLU'),
                 norm_eval: bool = False,
                 with_cp: bool = False,
                 init_cfg: Optional[Union[Dict, List[Dict]]] = None) -> None:
        layers_nums = 5 if with_last_layer else 4
        for index in out_indices:
            if index not in range(0, layers_nums):
                raise ValueError('the item in out_indices must in '
                                 f'range(0, 5). But received {index}')

        self.frozen_stages = frozen_stages
        if frozen_stages not in range(-1, layers_nums):
            raise ValueError('frozen_stages must be in range(-1, 5). '
                             f'But received {frozen_stages}')

        super().__init__(init_cfg)

        self.arch_setting = arch_setting
        self.widen_factor = widen_factor
        self.out_indices = out_indices
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.norm_eval = norm_eval
        self.with_cp = with_cp

        last_channels = 1024
        self.in_channels = 16 * stem_multiplier
        self.conv1 = ConvModule(
            in_channels=3,
            out_channels=self.in_channels,
            kernel_size=3,
            stride=2,
            padding=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

        self.layers = ModuleList()
        for channel, num_blocks, mutable_cfg in arch_setting:
            out_channels = round(channel * widen_factor)
            layer = self._make_layer(out_channels, num_blocks,
                                     copy.deepcopy(mutable_cfg))
            self.layers.append(layer)

        if with_last_layer:
            self.layers.append(
                ConvModule(
                    in_channels=self.in_channels,
                    out_channels=last_channels,
                    kernel_size=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))

    def _make_layer(self, out_channels: int, num_blocks: int,
                    mutable_cfg: Dict) -> Sequential:
        """Stack mutable blocks to build a layer for ShuffleNet V2.

        Note:
            Here we use ``module_kwargs`` to pass dynamic parameters such as
            ``in_channels``, ``out_channels`` and ``stride``
            to build the mutable.

        Args:
            out_channels (int): out_channels of the block.
            num_blocks (int): number of blocks.
            mutable_cfg (dict): Config of mutable.

        Returns:
            mmengine.model.Sequential: The layer made.
        """
        layers = []
        for i in range(num_blocks):
            stride = 2 if i == 0 else 1

            mutable_cfg.update(
                module_kwargs=dict(
                    in_channels=self.in_channels,
                    out_channels=out_channels,
                    stride=stride))
            layers.append(MODELS.build(mutable_cfg))
            self.in_channels = out_channels

        return Sequential(*layers)

    def _freeze_stages(self) -> None:
        """Freeze params not to update in the specified stages."""
        if self.frozen_stages >= 0:
            for param in self.conv1.parameters():
                param.requires_grad = False

        for i in range(self.frozen_stages):
            m = self.layers[i]
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

    def init_weights(self) -> None:
        """Init weights of ``SearchableShuffleNetV2``."""
        super().init_weights()

        if (isinstance(self.init_cfg, dict)
                and self.init_cfg['type'] == 'Pretrained'):
            # Suppress default init if use pretrained model.
            return

        for name, m in self.named_modules():
            if isinstance(m, nn.Conv2d):
                if 'conv1' in name:
                    normal_init(m, mean=0, std=0.01)
                else:
                    normal_init(m, mean=0, std=1.0 / m.weight.shape[1])
            elif isinstance(m, (_BatchNorm, nn.GroupNorm)):
                constant_init(m, val=1, bias=0.0001)
                if isinstance(m, _BatchNorm):
                    if m.running_mean is not None:
                        nn.init.constant_(m.running_mean, 0)

    def forward(self, x: Tensor) -> Tuple[Tensor, ...]:
        """Forward computation.

        Args:
            x (tensor): x contains input data for forward computation.
        """
        x = self.conv1(x)

        outs = []
        for i, layer in enumerate(self.layers):
            x = layer(x)
            if i in self.out_indices:
                outs.append(x)

        return tuple(outs)

    def train(self, mode: bool = True) -> None:
        """Set module status before forward computation."""
        super().train(mode)

        self._freeze_stages()
        if mode and self.norm_eval:
            for m in self.modules():
                if isinstance(m, _BatchNorm):
                    m.eval()



# Copyright (c) OpenMMLab. All rights reserved.
from mmrazor.registry import TASK_UTILS
from ..flops_params_counter import get_model_parameters_number
from .base_counter import BaseCounter


@TASK_UTILS.register_module()
class ReLUCounter(BaseCounter):
    """FLOPs/params counter for ReLU series activate function."""

    @staticmethod
    def add_count_hook(module, input, output):
        active_elements_count = output.numel()
        module.__flops__ += int(active_elements_count)
        module.__params__ += get_model_parameters_number(module)


@TASK_UTILS.register_module()
class PReLUCounter(ReLUCounter):
    pass


@TASK_UTILS.register_module()
class ELUCounter(ReLUCounter):
    pass


@TASK_UTILS.register_module()
class LeakyReLUCounter(ReLUCounter):
    pass


@TASK_UTILS.register_module()
class ReLU6Counter(ReLUCounter):
    pass



# Copyright (c) OpenMMLab. All rights reserved.
from .ab_loss import ABLoss
from .at_loss import ATLoss
from .cwd import ChannelWiseDivergence
from .dafl_loss import ActivationLoss, InformationEntropyLoss, OnehotLikeLoss
from .decoupled_kd import DKDLoss
from .factor_transfer_loss import FTLoss
from .fbkd_loss import FBKDLoss
from .kd_soft_ce_loss import KDSoftCELoss
from .kl_divergence import KLDivergence
from .l1_loss import L1Loss
from .l2_loss import L2Loss
from .ofd_loss import OFDLoss
from .relational_kd import AngleWiseRKD, DistanceWiseRKD
from .weighted_soft_label_distillation import WSLD

__all__ = [
    'ChannelWiseDivergence', 'KLDivergence', 'AngleWiseRKD', 'DistanceWiseRKD',
    'WSLD', 'L2Loss', 'ABLoss', 'DKDLoss', 'KDSoftCELoss', 'ActivationLoss',
    'OnehotLikeLoss', 'InformationEntropyLoss', 'FTLoss', 'ATLoss', 'OFDLoss',
    'L1Loss', 'FBKDLoss'
]



# Copyright (c) OpenMMLab. All rights reserved.
import copy

import numpy as np

from mmrazor.utils import SingleMutatorRandomSubnet


def crossover(random_subnet1: SingleMutatorRandomSubnet,
              random_subnet2: SingleMutatorRandomSubnet,
              prob: float = 0.5) -> SingleMutatorRandomSubnet:
    """Crossover in genetic algorithm.

    Args:
        random_subnet1 (SINGLE_MUTATOR_RANDOM_SUBNET): One of the subnets to
            crossover.
        random_subnet2 (SINGLE_MUTATOR_RANDOM_SUBNET): One of the subnets to
            crossover.
        prob (float): The probablity of getting choice from `random_subnet2`.
            Defaults to 0.5.

    Returns:
        SINGLE_MUTATOR_RANDOM_SUBNET: The result of crossover.
    """
    assert prob >= 0. and prob <= 1.,  \
        'The probability of crossover has to be between 0 and 1'
    crossover_subnet = copy.deepcopy(random_subnet1)
    for group_id, choice in random_subnet2.items():
        if np.random.random_sample() < prob:
            crossover_subnet[group_id] = choice
    return crossover_subnet



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.mutables.base_mutable import BaseMutable
from mmrazor.registry import MODELS
from .dynamic_mixins import DynamicBatchNormMixin


class _DynamicBatchNorm(_BatchNorm, DynamicBatchNormMixin):
    """Dynamic BatchNormxd OP.

    Note:
        Arguments for ``__init__`` of ``DynamicBatchNormxd`` is totally same as
        :obj:`torch.nn.BatchNormxd`.

    Attributes:
        mutable_attrs (ModuleDict[str, BaseMutable]): Mutable attributes,
            such as `num_features`. The key of the dict must in
            ``accepted_mutable_attrs``.
    """
    accepted_mutable_attrs = {'num_features'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.mutable_attrs: Dict[str, Optional[BaseMutable]] = nn.ModuleDict()

    @classmethod
    def convert_from(cls, module: _BatchNorm):
        """Convert a _BatchNorm module to a DynamicBatchNorm.

        Args:
            module (:obj:`torch.nn._BatchNorm`): The original BatchNorm module.
        """
        dynamic_bn = cls(
            num_features=module.num_features,
            eps=module.eps,
            momentum=module.momentum,
            affine=module.affine,
            track_running_stats=module.track_running_stats)

        return dynamic_bn

    def forward(self, input: Tensor) -> Tensor:
        """Forward of dynamic BatchNormxd OP."""
        self._check_input_dim(input)

        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:  # type: ignore
                self.num_batches_tracked = \
                    self.num_batches_tracked + 1  # type: ignore
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(
                        self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is
                                                           None)

        running_mean, running_var, weight, bias = self.get_dynamic_params()

        out = F.batch_norm(input, running_mean, running_var, weight, bias,
                           bn_training, exponential_average_factor, self.eps)

        # copy changed running statistics
        if self.training and self.track_running_stats:
            out_mask = self._get_num_features_mask()
            self.running_mean.masked_scatter_(out_mask, running_mean)
            self.running_var.masked_scatter_(out_mask, running_var)

        return out


@MODELS.register_module()
class DynamicBatchNorm1d(_DynamicBatchNorm):
    """Dynamic BatchNorm1d OP."""

    @property
    def static_op_factory(self):
        return nn.BatchNorm1d

    def _check_input_dim(self, input: Tensor) -> None:
        """Check if input dimension is valid."""
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError('expected 2D or 3D input (got {}D input)'.format(
                input.dim()))


@MODELS.register_module()
class DynamicBatchNorm2d(_DynamicBatchNorm):
    """Dynamic BatchNorm2d OP."""

    @property
    def static_op_factory(self):
        return nn.BatchNorm2d

    def _check_input_dim(self, input: Tensor) -> None:
        """Check if input dimension is valid."""
        if input.dim() != 4:
            raise ValueError('expected 4D input (got {}D input)'.format(
                input.dim()))


@MODELS.register_module()
class DynamicBatchNorm3d(_DynamicBatchNorm):
    """Dynamic BatchNorm3d OP."""

    @property
    def static_op_factory(self):
        return nn.BatchNorm3d

    def _check_input_dim(self, input: Tensor) -> None:
        """Check if input dimension is valid."""
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'.format(
                input.dim()))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch
from mmengine.model import BaseModule

from mmrazor.models.utils import get_module_device


class BaseGenerator(BaseModule):
    """The base class for generating images.

    Args:
        img_size (int): The size of generated image.
        latent_dim (int): The dimension of latent data.
        hidden_channels (int): The dimension of hidden channels.
        init_cfg (dict, optional): The config to control the initialization.
            Defaults to None.
    """

    def __init__(self,
                 img_size: int,
                 latent_dim: int,
                 hidden_channels: int,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg=init_cfg)
        self.img_size = img_size
        self.latent_dim = latent_dim
        self.hidden_channels = hidden_channels

    def process_latent(self,
                       latent_data: Optional[torch.Tensor] = None,
                       batch_size: int = 1) -> torch.Tensor:
        """Generate the latent data if the input is None. Put the latent data
        into the current gpu.

        Args:
            latent_data (torch.Tensor, optional): The latent data. Defaults to
                None.
            batch_size (int): The batch size of the latent data. Defaults to 1.
        """
        if isinstance(latent_data, torch.Tensor):
            assert latent_data.shape[1] == self.latent_dim, \
                'Second dimension of the input must be equal to "latent_dim",'\
                f'but got {latent_data.shape[1]} != {self.latent_dim}.'
            if latent_data.ndim == 2:
                batch_data = latent_data
            else:
                raise ValueError('The noise should be in shape of (n, c)'
                                 f'but got {latent_data.shape}')
        elif latent_data is None:
            assert batch_size > 0, \
                '"batch_size" should larger than zero when "latent_data" is '\
                f'None, but got {batch_size}.'
            batch_data = torch.randn((batch_size, self.latent_dim))

        # putting data on the right device
        batch_data = batch_data.to(get_module_device(self))
        return batch_data

    def forward(self) -> None:
        """Forward function."""
        raise NotImplementedError



# Copyright (c) OpenMMLab. All rights reserved.
import functools
from types import FunctionType, ModuleType
from typing import Callable

from mmengine.utils import import_modules_from_strings

from mmrazor.registry import TASK_UTILS
from .distill_delivery import DistillDelivery


@TASK_UTILS.register_module()
class MethodOutputsDelivery(DistillDelivery):
    """Delivery for intermediate results which are ``MethodType``'s outputs.

    Note:
        Different from ``FunctionType``, ``MethodType`` is the type of methods
        of class instances.

    Args:
        method_path (str): The name of the method whose output needs to be
            delivered.
        max_keep_data (int): The length limitation of the queue. Outputs from
            the source model are pushed in the queue in order.

    Examples:
        >>> from mmcls.models.utils import Augments

        >>> augments_cfg = dict(
        ...     type='BatchMixup', alpha=1., num_classes=10, prob=1.0)
        >>> augments = Augments(augments_cfg)
        >>> imgs = torch.randn(2, 3, 32, 32)
        >>> label = torch.randint(0, 10, (2,))

        >>> # Without ``MethodOutputsDeliver``, outputs of the teacher and the
        >>> # student are very likely to be different.
        >>> imgs_tea, label_tea = augments(imgs, label)
        >>> imgs_stu, label_stu = augments(imgs, label)
        >>> torch.equal(label_tea, label_stu)
        False
        >>> torch.equal(imgs_tea, imgs_stu)
        False

        >>> # Suppose we want to deliver outputs from the teacher to
        >>> # the student
        >>> delivery = MethodOutputsDeliver(
        ...     max_keep_data=1,
        ...     method_path='mmcls.models.utils.Augments.__call__')

        >>> delivery.override_data = False
        >>> with delivery:
        ...     imgs_tea, label_tea = augments(imgs, label)

        >>> delivery.override_data = True
        >>> with delivery:
        ...     imgs_stu, label_stu = augments(imgs, label)

        >>> torch.equal(label_tea, label_stu)
        True
        >>> torch.equal(imgs_tea, imgs_stu)
        True
    """

    def __init__(self, method_path: str, max_keep_data: int):
        super().__init__(max_keep_data)

        self._check_valid_path(method_path)
        module_path = self._get_module_path(method_path)
        try:
            module: ModuleType = import_modules_from_strings(module_path)
        except ImportError:
            raise ImportError(f'{module_path} is not imported correctly.')

        cls_name = self._get_cls_name(method_path)
        assert hasattr(module, cls_name), \
            f'{cls_name} is not in {module_path}.'

        imported_cls: type = getattr(module, cls_name)
        if not isinstance(imported_cls, type):
            raise TypeError(f'{cls_name} should be a type '
                            f'instance, but got {type(imported_cls)}')
        self.imported_cls = imported_cls

        method_name = self._get_method_name(method_path)
        assert hasattr(imported_cls, method_name), \
            f'{method_name} is not in {cls_name}.'
        self.method_name = method_name

        origin_method = getattr(imported_cls, method_name)
        # Before instantiation of a class, the type of a method of a class
        # is FunctionType
        if not isinstance(origin_method, FunctionType):
            raise TypeError(f'{method_name} should be a FunctionType  '
                            f'instance, but got {type(origin_method)}')
        self.origin_method = origin_method

    @staticmethod
    def _check_valid_path(method_path: str) -> None:
        """Check if the `method_path` is valid."""
        if not isinstance(method_path, str):
            raise TypeError(f'method_path should be a str instance, '
                            f'but got {type(method_path)}')

        assert len(method_path.split('.')) > 2, \
            'method_path must have at least one `.`'

    @staticmethod
    def _get_method_name(method_path: str) -> str:
        """Get the method name according to `method_path`."""
        return method_path.split('.')[-1]

    @staticmethod
    def _get_cls_name(method_path: str) -> str:
        """Get the class name corresponding to this method according to
        `method_path`."""
        return method_path.split('.')[-2]

    @staticmethod
    def _get_module_path(method_path: str) -> str:
        """Get the module name according to `method_path`."""
        return '.'.join(method_path.split('.')[:-2])

    def __enter__(self) -> None:
        """Enter the context manager.

        Wrap the origin method.
        """
        wrapped_method = self.deliver_wrapper(self.origin_method)
        setattr(self.imported_cls, self.method_name, wrapped_method)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager.

        Reset the origin method.
        """
        setattr(self.imported_cls, self.method_name, self.origin_method)

    def deliver_wrapper(self, origin_method: Callable) -> Callable:
        """Wrap the specific method to make the intermediate results of the
        model can be delivered."""

        @functools.wraps(origin_method)
        def wrap_method(*args, **kwargs):

            if self.override_data:
                assert not self.data_queue.empty(), 'pop from an empty queue'
                outputs = self.data_queue.get()
            else:
                assert not self.data_queue.full(), 'push into an full queue'
                outputs = origin_method(*args, **kwargs)
                self.data_queue.put(outputs)
            return outputs

        return wrap_method



_base_ = ['./spos_shufflenet_supernet_8xb128_in1k.py']

# FIXME: you may replace this with the mutable_cfg searched by yourself
fix_subnet = 'https://download.openmmlab.com/mmrazor/v1/spos/spos_shufflenetv2_subnet_8xb128_in1k_flops_0.33M_acc_73.87_20220715-aa94d5ef_subnet_cfg_v1.yaml'  # noqa: E501

model = dict(fix_subnet=fix_subnet)

find_unused_parameters = False



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from queue import Queue
from typing import Callable


# TODO: Support overriding part of the outputs of a function or method
class DistillDelivery(metaclass=ABCMeta):
    """Base class for deliveries for distillation.

    DistillDelivery is a context manager used to override function(method)
    outputs during teacher(student) forward.

    A delivery can only handle one function or method. Some algorithms may use
    multiple deliveries, which can be managed uniformly using
    ``DistillDeliverManager``.

    Args:
        max_keep_data (int): The length limitation of the queue, should be
            larger than the execute times of the function or method. Defaults
            to 1.

    Notes:
        If a function (method) is executed more than once during the forward
        of the source model, all the outputs of this function (method) will be
        used to override function (method) outputs from the target model.

        If a function or method is executed more than once during the forward
        of the target model, its' outputs from the source model are pushed
        into the queue in order.
    """

    def __init__(self, max_keep_data: int = 1) -> None:

        self._override_data = False
        self.data_queue: Queue = Queue(maxsize=max_keep_data)
        self.max_keep_data = max_keep_data

    @property
    def override_data(self) -> bool:
        """bool: indicate whether to override the data with the recorded data.
        """
        return self._override_data

    @override_data.setter
    def override_data(self, override: bool) -> None:
        """Set the override_data property to this delivery.

        If the `override_data` of a deliver is False, the deliver will record
        and keep the origin data. If the current_mode of a deliver is True, the
        deliver will override the origin data with the recorded data.
        """
        self._override_data = override

    @abstractmethod
    def deliver_wrapper(self, origin: Callable) -> Callable:
        """Wrap the specific object to make the intermediate results of the
        model can be delivered."""

    @abstractmethod
    def __enter__(self) -> None:
        """Enter the context manager."""

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager."""



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Optional, Tuple

import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn.modules import GroupNorm
from torch.nn.modules.batchnorm import _BatchNorm
from torch.nn.modules.instancenorm import _InstanceNorm

from mmrazor.models.mutables.mutable_channel import MutableChannel
from mmrazor.registry import MODELS
from .base import ChannelDynamicOP


class DynamicConv2d(nn.Conv2d, ChannelDynamicOP):
    """Applies a 2D convolution over an input signal composed of several input
    planes according to the `mutable_in_channels` and `mutable_out_channels`
    dynamically.

    Args:
        in_channels_cfg (Dict): Config related to `in_channels`.
        out_channels_cfg (Dict): Config related to `out_channels`.
    """
    accepted_mutables = {'mutable_in_channels', 'mutable_out_channels'}

    def __init__(self, in_channels_cfg, out_channels_cfg, *args, **kwargs):
        super(DynamicConv2d, self).__init__(*args, **kwargs)

        in_channels_cfg_ = copy.deepcopy(in_channels_cfg)
        in_channels_cfg_.update(dict(num_channels=self.in_channels))
        self.mutable_in_channels = MODELS.build(in_channels_cfg_)

        out_channels_cfg_ = copy.deepcopy(out_channels_cfg)
        out_channels_cfg_.update(dict(num_channels=self.out_channels))
        self.mutable_out_channels = MODELS.build(out_channels_cfg_)

        assert isinstance(self.mutable_in_channels, MutableChannel)
        assert isinstance(self.mutable_out_channels, MutableChannel)
        # TODO
        # https://pytorch.org/docs/stable/_modules/torch/nn/modules/conv.html#Conv2d
        assert self.padding_mode == 'zeros'

    @property
    def mutable_in(self) -> MutableChannel:
        """Mutable `in_channels`."""
        return self.mutable_in_channels

    @property
    def mutable_out(self) -> MutableChannel:
        """Mutable `out_channels`."""
        return self.mutable_out_channels

    def forward(self, input: Tensor) -> Tensor:
        """Slice the parameters according to `mutable_in_channels` and
        `mutable_out_channels`, and forward."""
        groups = self.groups
        if self.groups == self.in_channels == self.out_channels:
            groups = input.size(1)
        weight, bias = self._get_dynamic_params()

        return F.conv2d(input, weight, bias, self.stride, self.padding,
                        self.dilation, groups)

    def _get_dynamic_params(self) -> Tuple[Tensor, Optional[Tensor]]:
        in_mask = self.mutable_in_channels.current_mask.to(self.weight.device)
        out_mask = self.mutable_out_channels.current_mask.to(
            self.weight.device)

        if self.groups == 1:
            weight = self.weight[out_mask][:, in_mask]
        elif self.groups == self.in_channels == self.out_channels:
            # depth-wise conv
            weight = self.weight[out_mask]
        else:
            raise NotImplementedError(
                'Current `ChannelMutator` only support pruning the depth-wise '
                '`nn.Conv2d` or `nn.Conv2d` module whose group number equals '
                f'to one, but got {self.groups}.')

        bias = self.bias[out_mask] if self.bias is not None else None

        return weight, bias

    def to_static_op(self) -> nn.Conv2d:
        assert self.mutable_in.is_fixed and self.mutable_out.is_fixed

        weight, bias, = self._get_dynamic_params()
        groups = self.groups
        if groups == self.in_channels == self.out_channels:
            groups = self.mutable_in.current_mask.sum().item()
        out_channels = weight.size(0)
        in_channels = weight.size(1) * groups

        static_conv2d = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            padding_mode=self.padding_mode,
            dilation=self.dilation,
            groups=groups,
            bias=True if bias is not None else False)

        static_conv2d.weight = nn.Parameter(weight)
        if bias is not None:
            static_conv2d.bias = nn.Parameter(bias)

        return static_conv2d


class DynamicLinear(nn.Linear, ChannelDynamicOP):
    """Applies a linear transformation to the incoming data according to the
    `mutable_in_features` and `mutable_out_features` dynamically.

    Args:
        in_features_cfg (Dict): Config related to `in_features`.
        out_features_cfg (Dict): Config related to `out_features`.
    """
    accepted_mutables = {'mutable_in_features', 'mutable_out_features'}

    def __init__(self, in_features_cfg, out_features_cfg, *args, **kwargs):
        super(DynamicLinear, self).__init__(*args, **kwargs)

        in_features_cfg_ = copy.deepcopy(in_features_cfg)
        in_features_cfg_.update(dict(num_channels=self.in_features))
        self.mutable_in_features = MODELS.build(in_features_cfg_)

        out_features_cfg_ = copy.deepcopy(out_features_cfg)
        out_features_cfg_.update(dict(num_channels=self.out_features))
        self.mutable_out_features = MODELS.build(out_features_cfg_)

    @property
    def mutable_in(self):
        """Mutable `in_features`."""
        return self.mutable_in_features

    @property
    def mutable_out(self):
        """Mutable `out_features`."""
        return self.mutable_out_features

    def forward(self, input: Tensor) -> Tensor:
        """Slice the parameters according to `mutable_in_features` and
        `mutable_out_features`, and forward."""
        in_mask = self.mutable_in_features.current_mask.to(self.weight.device)
        out_mask = self.mutable_out_features.current_mask.to(
            self.weight.device)

        weight = self.weight[out_mask][:, in_mask]
        bias = self.bias[out_mask] if self.bias is not None else None

        return F.linear(input, weight, bias)

    # TODO
    def to_static_op(self) -> nn.Module:
        return self


class DynamicBatchNorm(_BatchNorm, ChannelDynamicOP):
    """Applies Batch Normalization over an input according to the
    `mutable_num_features` dynamically.

    Args:
        num_features_cfg (Dict): Config related to `num_features`.
    """
    accepted_mutables = {'mutable_num_features'}

    def __init__(self, num_features_cfg, *args, **kwargs):
        super(DynamicBatchNorm, self).__init__(*args, **kwargs)

        num_features_cfg_ = copy.deepcopy(num_features_cfg)
        num_features_cfg_.update(dict(num_channels=self.num_features))
        self.mutable_num_features = MODELS.build(num_features_cfg_)

    @property
    def mutable_in(self):
        """Mutable `num_features`."""
        return self.mutable_num_features

    @property
    def mutable_out(self):
        """Mutable `num_features`."""
        return self.mutable_num_features

    def forward(self, input: Tensor) -> Tensor:
        """Slice the parameters according to `mutable_num_features`, and
        forward."""
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:  # type: ignore
                self.num_batches_tracked = \
                    self.num_batches_tracked + 1  # type: ignore
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(
                        self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is
                                                           None)

        if self.affine:
            out_mask = self.mutable_num_features.current_mask.to(
                self.weight.device)
            weight = self.weight[out_mask]
            bias = self.bias[out_mask]
        else:
            weight, bias = self.weight, self.bias

        if self.track_running_stats:
            out_mask = self.mutable_num_features.current_mask.to(
                self.running_mean.device)
            running_mean = self.running_mean[out_mask] \
                if not self.training or self.track_running_stats else None
            running_var = self.running_var[out_mask] \
                if not self.training or self.track_running_stats else None
        else:
            running_mean, running_var = self.running_mean, self.running_var

        return F.batch_norm(input, running_mean, running_var, weight, bias,
                            bn_training, exponential_average_factor, self.eps)

    # TODO
    def to_static_op(self) -> nn.Module:
        return self


class DynamicInstanceNorm(_InstanceNorm, ChannelDynamicOP):
    """Applies Instance Normalization over an input according to the
    `mutable_num_features` dynamically.

    Args:
        num_features_cfg (Dict): Config related to `num_features`.
    """
    accepted_mutables = {'mutable_num_features'}

    def __init__(self, num_features_cfg, *args, **kwargs):
        super(DynamicInstanceNorm, self).__init__(*args, **kwargs)

        num_features_cfg_ = copy.deepcopy(num_features_cfg)
        num_features_cfg_.update(dict(num_channels=self.num_features))
        self.mutable_num_features = MODELS.build(num_features_cfg_)

    @property
    def mutable_in(self):
        """Mutable `num_features`."""
        return self.mutable_num_features

    @property
    def mutable_out(self):
        """Mutable `num_features`."""
        return self.mutable_num_features

    def forward(self, input: Tensor) -> Tensor:
        """Slice the parameters according to `mutable_num_features`, and
        forward."""
        if self.affine:
            out_mask = self.mutable_num_features.current_mask.to(
                self.weight.device)
            weight = self.weight[out_mask]
            bias = self.bias[out_mask]
        else:
            weight, bias = self.weight, self.bias

        if self.track_running_stats:
            out_mask = self.mutable_num_features.current_mask.to(
                self.running_mean.device)
            running_mean = self.running_mean[out_mask]
            running_var = self.running_var[out_mask]
        else:
            running_mean, running_var = self.running_mean, self.running_var

        return F.instance_norm(input, running_mean, running_var, weight, bias,
                               self.training or not self.track_running_stats,
                               self.momentum, self.eps)

    # TODO
    def to_static_op(self) -> nn.Module:
        return self


class DynamicGroupNorm(GroupNorm, ChannelDynamicOP):
    """Applies Group Normalization over a mini-batch of inputs according to the
    `mutable_num_channels` dynamically.

    Args:
        num_channels_cfg (Dict): Config related to `num_channels`.
    """
    accepted_mutables = {'mutable_num_features'}

    def __init__(self, num_channels_cfg, *args, **kwargs):
        super(DynamicGroupNorm, self).__init__(*args, **kwargs)

        num_channels_cfg_ = copy.deepcopy(num_channels_cfg)
        num_channels_cfg_.update(dict(num_channels=self.num_channels))
        self.mutable_num_channels = MODELS.build(num_channels_cfg_)

    @property
    def mutable_in(self):
        """Mutable `num_channels`."""
        return self.mutable_num_channels

    @property
    def mutable_out(self):
        """Mutable `num_channels`."""
        return self.mutable_num_channels

    def forward(self, input: Tensor) -> Tensor:
        """Slice the parameters according to `mutable_num_channels`, and
        forward."""
        if self.affine:
            out_mask = self.mutable_num_channels.current_mask.to(
                self.weight.device)
            weight = self.weight[out_mask]
            bias = self.bias[out_mask]
        else:
            weight, bias = self.weight, self.bias

        return F.group_norm(input, self.num_groups, weight, bias, self.eps)

    # TODO
    def to_static_op(self) -> nn.Module:
        return self
